# cython: embedsignature=True
# cython: profile=False

import math
import random

cimport cython

from cpython.version cimport PY_MAJOR_VERSION, PY_MINOR_VERSION
from cpython.bytes cimport PyBytes_AsString, PyBytes_FromStringAndSize
from cpython cimport PyBytes_Check, PyUnicode_Check
from libc.stdlib cimport qsort, realloc, malloc, calloc, free
from libc.stdint cimport UINT32_MAX
from libc.string cimport memcpy, memmove
from libc.errno cimport errno
from libc.math cimport floor
from posix.types cimport off_t
from posix.mman cimport mmap, munmap, shm_open, shm_unlink
from posix.mman cimport MAP_SHARED, PROT_READ, PROT_WRITE
from posix.stat cimport S_IRUSR, S_IWUSR
from posix.fcntl cimport O_CREAT, O_RDWR, O_RDONLY
from posix.unistd cimport ftruncate


cdef bytes force_bytes(object s, encoding="ascii"):
    """convert string or unicode object to bytes, assuming
    ascii encoding.
    """
    if s is None:
        return None
    elif PyBytes_Check(s):
        return s
    elif PyUnicode_Check(s):
        return s.encode(encoding)
    else:
        raise TypeError("Argument must be string, bytes or unicode.")

cdef force_str(object s, encoding="ascii"):
    """Return s converted to str type of current Python
    (bytes in Py2, unicode in Py3)"""
    if s is None:
        return None
    if PY_MAJOR_VERSION < 3:
        return s
    elif PyBytes_Check(s):
        return s.decode(encoding)
    else:
        # assume unicode
        return s

#####################################################
## numpy import
## both import and cimport are necessary
import numpy
cimport numpy
DTYPE_INT = numpy.int
ctypedef numpy.int_t DTYPE_INT_t
DTYPE_FLOAT = numpy.float
ctypedef numpy.float_t DTYPE_FLOAT_t

def getSegmentSize():
    '''return size of coordinate and size of a segment.'''
    return sizeof(Segment) // 2, sizeof(Segment)

# min/max are not optimized, so declare them as C functions
# declare as signed comparisons as a Position might be negative
@cython.profile(False)
cdef inline PositionDifference lmin(PositionDifference a, PositionDifference b) nogil:
    if a < b:
        return a
    return b

@cython.profile(False)
cdef inline PositionDifference lmax(PositionDifference a, PositionDifference b) nogil:
    if a > b:
        return a
    return b

@cython.profile(False)
cdef inline double dmin(double a, double b) nogil:
    if a < b:
        return a
    return b

@cython.profile(False)
cdef inline double dmax(double a, double b) nogil:
    if a > b:
        return a
    return b

@cython.profile(False)
cdef inline PositionDifference segment_overlap(Segment a, Segment b) nogil:
    """return number of positions overlapping between to Segments."""
    return lmax(0,
                <PositionDifference>lmin( a.end, b.end) -
                <PositionDifference>lmax(a.start, b.start))

@cython.profile(False)
cdef inline PositionDifference range_overlap(
    Position astart, Position aend, Position bstart, Position bend) nogil:
    return lmax(0, 
                <PositionDifference>lmin( aend, bend) - 
                <PositionDifference>lmax(astart, bstart))

@cython.profile(False)
cdef inline PositionDifference segment_overlap_raw(Segment a, Segment b) nogil:
    return <PositionDifference>lmin(a.end, b.end) - \
        <PositionDifference>lmax(a.start, b.start)

@cython.profile(False)
cdef inline PositionDifference segment_length(Segment a) nogil:
    return <PositionDifference>a.end - <PositionDifference>a.start

# trick to permit const void * in function definitions
cdef extern from *:
    ctypedef void * const_void_ptr "const void*"

@cython.profile(False)
cdef int cmpSegments(const_void_ptr s1, const_void_ptr s2) nogil:
    return <PositionDifference>(<Segment *>s1).start - \
        <PositionDifference>(<Segment *>s2).start

@cython.profile(False)
cdef int cmpSegmentsStartAndEnd(const_void_ptr s1, const_void_ptr s2) nogil:
    cdef int x
    x = <PositionDifference>(<Segment *>s1).start - \
        <PositionDifference>(<Segment *>s2).start
    if x != 0:
        return x
    return <PositionDifference>(<Segment *>s1).end - \
        <PositionDifference>(<Segment *>s2).end

@cython.profile(False)
cdef int cmpPosition(const_void_ptr s1, const_void_ptr s2) nogil:
    return (<Position*>s1)[0] - (<Position*>s2)[0]

@cython.profile(False)
cdef int cmpDouble( const_void_ptr s1, const_void_ptr s2 ) nogil:
    return <int>((<double*>s1)[0] - (<double*>s2)[0])

# Flag for Segment list status
DEF SEG_NORMALIZED = 1
DEF SEG_SHARED = 2
DEF SEG_SLAVE = 4

cdef class SegmentList:
    '''list of segments.

    A list of (non-overlapping) segments.

    The intersection algorithms assume that segments
    are non-overlapping and sorted by start.

    Call normalize to merge overlapping segments.
    '''
    def __init__(self, 
                 int allocate=0,
                 SegmentList clone=None,
                 iter=None,
                 normalize=False,
                 SegmentList share=None,
                 unreduce=None):
        '''create empty list of segments.
        
        Creating is a lazy operation, no memory is allocated unless
        explitely given. A SegmentList might be initialized using
        the following options:

        If *allocate* is set, create an empty list, but allocate slots

        If *clone* is given, create a copy an existing list.

        If *iter* is given, initialize list from iterable.

        If *share* is set, the list will be a slave. It will share
        memory with a _master_, which has been set up by calling the 
        :meth:`share` method. Memory access for the slave is read-only.

        If *unreduce* is given, the list will be re-constituted
        from a pickled data representation (see __reduce__). Note that
        if a _master_ has been pickled, the reconstituted object will be
        a _slave_.

        If *normalize* is set, the list will be normalized.
        '''
        cdef char * p
        cdef long idx, nsegments
        cdef void * retval
        cdef int error_number
        cdef char * buffer

        # initialize empty list
        self.segments = NULL
        self.nsegments = 0
        self.allocated = 0

        # The state of a segment list can be:
        # 0. Empty and private
        # allocated >= 0, shared_fd = -1, nsegments = 0
        # 1. Private
        # allocated > 0, shared_fd = -1, nsegments >= 0
        # Only segment lists with segments can be shared
        # 2. Shared, through segment list, master
        # allocated = 0, shared_fd > 0, nsegments > 0
        # 3. Shared, through segment list, slave
        # allocated = 0, shared_fd = -1, nsegments > 0
        # 4. Shared, through IntervalCollection, master
        # allocated = 0, shared_fd = -1, nsegments > 0
        # 5. Shared, through IntervalCollection, slave
        # allocated = 0, shared_fd = -1, nsegments > 0

        # an empty list is normalized
        self.flag = 1
        self.chunk_size = 10000

        # no sharing
        self.shared_fd = -1
        self.key = None

        self.is_shared = False
        self.is_slave = False

        # initialize list from shared memory
        # allocated is 0
        if share != None:
            self.nsegments = share.nsegments
            if self.nsegments > 0:
                assert share.shared_fd >= 0, "sharing from a non-shared SegmentList"

                self.allocated = 0
                self.flag = share.flag
                self.is_shared = True
                self.is_slave = True
                self.chunk_size = share.chunk_size
                retval = <Segment*>mmap(NULL, 
                                        self.nsegments * sizeof(Segment),
                                        PROT_READ, 
                                        MAP_SHARED, 
                                        share.shared_fd, 
                                        0)

                error_number = errno

                if retval == <void*>-1:
                    raise ValueError(
                        "could not read list from shared segment - error=%i" %
                        error_number)
                self.segments = <Segment *>retval

        # create from pickled representation
        # if a shared object is pickled, it will be re-constituted 
        # as a slave list.
        elif unreduce:
            self.nsegments, self.allocated, self.flag, \
                self.chunk_size, self.key = unreduce[:5]
            if type(unreduce[5]) == int:
                # shared memory
                shared_fd = unreduce[5]
                
                retval = mmap(NULL, 
                              self.nsegments * sizeof(Segment),
                              PROT_READ, 
                              MAP_SHARED, 
                              shared_fd, 
                              0)

                error_number = errno

                if retval == <void*>-1:
                    raise ValueError(
                        "could not unpickle as slave - error=%i" % error_number)

                self.segments = <Segment *>retval

                # mark memory as slave - nothing allocated and no shared_fd
                self.allocated = 0
                self.shared_fd = -1
                self.is_shared = True
                self.is_slave = True
            else:
                p = PyBytes_AsString(unreduce[5])
                self.segments = <Segment*>malloc(self.nsegments * sizeof(Segment))
                memcpy(self.segments, p, cython.sizeof(Position) * 2 * self.nsegments)

        # clone from another list
        elif clone != None:
            self.nsegments = self.allocated = clone.nsegments
            self.segments = <Segment*>calloc(clone.nsegments, sizeof( Segment))
            memcpy(self.segments,
                   clone.segments,
                   clone.nsegments * sizeof(Segment))
            self.flag = clone.flag & SEG_NORMALIZED
            
        # create from an iterable
        elif iter:
            a = tuple(iter)
            nsegments = len(a)
            self.nsegments = self.allocated = nsegments
            self.segments = <Segment*>calloc(self.nsegments, sizeof(Segment))
            idx = 0
            for start, end in a:
                self.segments[idx] = Segment(start, end)
                idx += 1
            self.flag = 0

        # allocated memory, list remains empty
        elif allocate:
            self.allocated = allocate
            self.segments = <Segment*>calloc(allocate, sizeof(Segment))

        if normalize:
            self.normalize()

    def __reduce__(self):
        '''pickling function - returns class contents as a tuple.'''

        cdef bytes data

        if self.shared_fd >= 0:
            return (buildSegmentList, (self.nsegments, 
                                       self.allocated, 
                                       self.flag, 
                                       self.chunk_size, 
                                       self.key,
                                       self.shared_fd))

        else:
            data = PyBytes_FromStringAndSize(
                <char*>self.segments, \
                self.nsegments * cython.sizeof(Position) * 2)
        
            return (buildSegmentList, (self.nsegments, 
                                       self.allocated, 
                                       self.flag, 
                                       self.chunk_size, 
                                       self.key,
                                       data))

    cpdef share(self, key):
        '''share data as a memory map.
        
        The data is shared under 'key'.

        Returns file descriptor of shared memory.
        '''
        assert self.flag & SEG_NORMALIZED, "sharing in non-normalized list"
        assert key.startswith( "/" )
        assert self.shared_fd == -1, "SegmentList already shared"
        cdef int error

        if self.nsegments == 0:
            # do not share empty objects
            return

        cdef int fd
        key = force_bytes(key)
        fd = shm_open(key, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR)
        if fd == -1:
            error = errno
            raise OSError( "could not create shared memory at %s; ERRNO=%i" % (key, error ))

        cdef off_t nbytes = sizeof(Segment) * self.nsegments
        # save file descriptor
        self.shared_fd = fd
        if ftruncate(fd, nbytes) == -1:
            error = errno
            raise OSError( "could not resize memory at %s; ERRNO=%i" % (key, error) )
        
        # open as memory map
        cdef void * p
        p = mmap(NULL, nbytes,
                 PROT_READ | PROT_WRITE, 
                 MAP_SHARED, 
                 fd, 
                 0);
        
        if p == <void*>-1:
            raise ValueError("could not create memory mapped file" )
        
        # copy data to shared memory location
        memcpy( p, self.segments, nbytes)
        self.key = key
        
        # free allocated private memory
        self.allocated = 0
        free(self.segments)

        self.segments = <Segment *>p
        self.is_shared = True
        
    def unshare(self):
        '''take ownership of shared data.
        '''
        # empty lists are not shared
        if self.nsegments == 0:
            return

        cdef int fd
        assert self.is_shared
        assert not self.is_slave
        
        if self.shared_fd == -1:
            raise ValueError("unsharing unshared SegmentList")

        fd = shm_unlink(self.key)
        if fd == -1:
            raise OSError( "could not unlink shared memory" )

        # copy data back to unshared memory
        cdef Segment * s
        cdef off_t nbytes = sizeof(Segment) * self.nsegments
        s = <Segment *>malloc(nbytes)
        if s == NULL:
            raise ValueError( "could not allocate memory when unsharing" )
        
        memcpy(s, self.segments, nbytes)
        
        # unmap
        if munmap(self.segments, nbytes) == -1:
            raise ValueError("error while unmapping memory")
        
        self.segments = s
        self.allocated = self.nsegments
        self.shared_fd = -1
        self.is_shared = False

    cdef void fromMMAP(self):
        '''retrieve data from mmapped location to private copy.
        '''
        # copy data back to unshared memory
        cdef Segment * s
        cdef off_t nbytes = sizeof(Segment) * self.nsegments
    
        if self.nsegments == 0:
            return

        s = <Segment *>malloc(nbytes)
        if s == NULL:
            raise ValueError( "could not allocate memory when unsharing" )
        
        memcpy(s, self.segments, nbytes)
        
        self.segments = s
        self.allocated = self.nsegments
        self.is_shared = False
        self.is_slave = False

    cdef off_t toMMAP(self, 
                      void * mmap, 
                      int shared_fd, 
                      off_t offset):
        '''move contents of segment list to shared memory 
        at *mmap* with *offset*.
        '''
        if self.shared_fd != -1: 
            raise ValueError("can not move shared segmentlist to mmap")

        if self.nsegments == 0:
            # do not move empty segment lists
            return offset
        
        cdef off_t nbytes = sizeof(Segment) * self.nsegments
        cdef Segment * p = <Segment *>mmap + offset

        # copy data
        memcpy(p, self.segments, nbytes)

        # free allocated private memory
        self.allocated = 0
        free(self.segments)
        self.segments = p

        self.is_shared = True
        self.is_slave = True

        return offset + self.nsegments

    cpdef sort(self):
        '''sort segments.'''
        if self.nsegments == 0:
            return

        qsort(<void*>self.segments,
              self.nsegments,
              sizeof(Segment),
              &cmpSegments)

    cpdef SegmentList extend(self, SegmentList other):
        '''extend a list of segments with segments in other list.

        The list will not be normalized automatically - call
        :meth:`SegmentList.normalize`.
        '''
        cdef size_t new_size
        new_size = other.nsegments + self.nsegments
        if self.allocated == 0:
            self.segments = <Segment*>malloc( new_size * sizeof( Segment ) )
            if not self.segments: 
                raise MemoryError( "out of memory when allocation %i bytes" % sizeof( new_size * sizeof( Segment ) ))
            self.allocated = new_size
        elif new_size >= self.allocated:
            self.segments = <Segment*>realloc( <void *>self.segments, new_size * sizeof( Segment ) )
            if not self.segments: 
                raise MemoryError( "out of memory when allocation %i bytes" % sizeof( new_size * sizeof( Segment ) ))
            self.allocated = new_size
        memcpy( &self.segments[self.nsegments],
                 other.segments,
                 other.nsegments * sizeof(Segment ) )
        self.nsegments = new_size
        self.flag = 0
        return self

    cdef _add(self, Segment segment):
        '''add a new segment.

        The list will not be normalized automatically - call
        :meth:`SegmentList.normalize`.
        '''

        if self.allocated == 0:
            self.segments = <Segment*>malloc(self.chunk_size * sizeof(Segment))
            if not self.segments:
                raise MemoryError("out of memory when allocation %i bytes" %
                                  sizeof( self.chunk_size * sizeof(Segment)))
            self.allocated = self.chunk_size
        elif self.nsegments == self.allocated:
            self.allocated *= 2
            self.segments = <Segment*>realloc(self.segments,
                                              self.allocated * sizeof(Segment))
            if not self.segments:
                raise MemoryError(
                    "out of memory when allocation %i bytes" %
                    sizeof(self.allocated * sizeof(Segment)))

        self.segments[self.nsegments] = segment
        self.nsegments += 1
        self.flag = 0

    cpdef add(self, Position start, Position end):
        cdef Segment segment
        assert start <= end, "attempting to add invalid segment %i-%i" % (start, end)
        segment = Segment(start, end)
        self._add(segment)

    cpdef trim_ends( self, Position pos, Position size, int forward ):
        '''trim segment list by removing *size* nucleotides from
        the segment that includes *pos*.
        
        The flag *forward* gives the direction.
        '''
        if self.nsegments == 0: return

        assert self.flag & SEG_NORMALIZED, "trimming in non-normalized list"

        cdef int idx
        cdef Position l
        cdef PositionDifference s = size
        cdef Segment other, seg

        assert self.sum() > s, "trimming more than the total length (%i < %i)" % (self.sum(), s)
        #print "at start=%i, removing %i" % (self.sum(), s)

        other = Segment( pos, pos + 1)
        idx = self._getInsertionPoint( other )

        # wrap around
        if idx == self.nsegments: idx = 0
        if idx < 0: idx = self.nsegments - 1

        if forward:
            while s > 0:
                seg = self.segments[idx]
                l = segment_length(seg)
                if segment_length(seg) < s:
                    self.segments[idx] = Segment(0,0)
                    s -= l
                else:
                    self.segments[idx] = Segment( seg.start + s, seg.end )
                    s = 0

                idx += 1
                if idx == self.nsegments: idx = 0
        else:
            while s > 0:
                seg = self.segments[idx]
                l = segment_length( seg )
                if segment_length( seg ) < s:
                    self.segments[idx] = Segment(0,0)
                    s -= l
                else:
                    self.segments[idx] = Segment( seg.start, <PositionDifference>seg.end - s )
                    s = 0

                idx -= 1
                if idx < 0: idx = self.nsegments - 1

        #print "at end=%i, removing %i" % (self.sum(), s)

    cdef _resize(self, int nsegments):
        '''resize segment list to nsegments.'''
        # do not free all if there are no segments
        if nsegments == 0:
            nsegments = 1

        assert nsegments >= self.nsegments, "resizing will loose segments"
        
        if self.allocated == 0:
            self.segments = <Segment*>malloc(nsegments * sizeof(Segment))
            if not self.segments:
                raise MemoryError("out of memory when allocation %i bytes" %
                                  sizeof(nsegments * sizeof(Segment)))
        else:
            self.segments = <Segment*>realloc(self.segments,
                                              nsegments * sizeof(Segment))
            if not self.segments:
                raise MemoryError("out of memory when allocation %i bytes" %
                                  sizeof(self.nsegments * sizeof(Segment)))

        self.allocated = nsegments

    cdef insert(self, int idx, Segment seg):
        '''insert Segment *seg* at position *idx*'''
        if idx < 0:
            raise ValueError("only positive indices accepted (%i)" % idx)
        
        idx = max(idx, self.nsegments)
        if self.allocated == self.nsegments:
            self._resize(self.allocated * 2)

        memmove(&self.segments[idx+1],
                &self.segments[idx],
                sizeof(Segment) * (self.nsegments - idx))

        self.nsegments += 1
        self.segments[idx] = seg

    cpdef trim(self, Position pos, Position size):
        '''trim segment list by removing *size* nucleotides
        starting at pos.

        If (pos, pos+size) is fully within a segment, this segment
        will be split.

        '''
        if self.nsegments == 0: return

        assert self.flag & SEG_NORMALIZED, "trimming in non-normalized list"

        cdef int idx
        cdef PositionDifference l
        cdef PositionDifference s = size
        cdef PositionDifference p = pos
        cdef Segment other, seg

        assert self.sum() > s, "trimming more than the total length (%i < %i)" % (self.sum(), s)

        p = lmax( p, self.segments[0].start )
        p = lmin( p, self.segments[self.nsegments-1].end )

        other = Segment( p, p + 1)
        idx = self._getInsertionPoint( other )
        # wrap around
        if idx == self.nsegments:
            idx = 0
            p = self.segments[0].start
        
        #assert p >= self.segments[idx].start, "pos=%i, %i, %s, %s" % (p, idx, str(self.segments[idx]), str(self.asList()))
        #assert p < self.segments[idx].end, "pos=%i, %i, %s, %s" % (p, idx, str(self.segments[idx]), str(self.asList()))

        while s > 0:
            seg = self.segments[idx]
            
            l = <PositionDifference>self.segments[idx].end - p

            if s < l: 
                if p == self.segments[idx].start:
                    # truncate from left and stop
                    self.segments[idx].start = p + s
                    break
                else:
                    # remove from middle and stop
                    self.insert( idx + 1, Segment( p + s, self.segments[idx].end) )
                    self.segments[idx].end = p
                    break
            else:
                self.segments[idx].end = p
                s -= l

            idx += 1

            # wrap around
            if idx == self.nsegments: idx = 0
            p = self.segments[idx].start

        #print "at end=%i, removing %i" % (self.sum(), s)

    cpdef normalize(self):
        '''merge all overlapping segments and remove empty segments.

        Adjacent segments are not merged.

        This function works in-place.
        '''

        cdef int idx, insertion_idx
        cdef Position max_end
        if self.nsegments == 0:
            self.flag = 1
            return

        self.sort()

        insertion_idx = 0

        # skip to first non-empty segment
        idx = 0
        while idx < self.nsegments and self.segments[idx].start == self.segments[idx].end: 
            idx += 1

        # only empty segments - empty list, but declare normalized
        if idx == self.nsegments:
            self.nsegments = 0
            self.flag = 1
            return

        self.segments[insertion_idx].start = self.segments[idx].start
        max_end = self.segments[idx].end
        
        while idx < self.nsegments:

            # skip empty
            if self.segments[idx].start == self.segments[idx].end: 
                idx += 1
                continue
            # no overlap - save segment
            if self.segments[idx].start >= max_end:
                self.segments[insertion_idx].end = max_end
                insertion_idx += 1
                assert insertion_idx < self.nsegments, \
                    "insertion_idx out of range: %i >= %i" % \
                    (insertion_idx, self.nsegments)
                self.segments[insertion_idx].start = self.segments[idx].start

            max_end = lmax(self.segments[idx].end, max_end)
            idx += 1

        # store final segment
        self.segments[insertion_idx].end = max_end

        # truncate array
        insertion_idx += 1
        self.nsegments = insertion_idx
        self._resize(self.nsegments)
        self.flag = 1

    cpdef merge(self, PositionDifference distance):
        '''merge all overlapping segments and remove empty segments.

        This function is a generalization of the merge method and
        will merge adjacent and/or neighbouring segments.

        If *distance* = 0, adjacent segments will be merged. 

        If*distance* = n, neighbouring segments with a separation of at most
        *n* bases will be merged.

        This function works in-place.
        '''

        cdef PositionDifference max_end
        cdef int idx, insertion_idx
        if self.nsegments == 0:
            self.flag = 1
            return

        self.sort()

        insertion_idx = 0

        # skip to first non-empty segment
        idx = 0
        while idx < self.nsegments and self.segments[idx].start == self.segments[idx].end: 
            idx += 1

        # only empty segments - empty list, but declare normalized
        if idx == self.nsegments:
            self.nsegments = 0
            self.flag = 1
            return

        self.segments[insertion_idx].start = self.segments[idx].start
        max_end = self.segments[idx].end
        
        while idx < self.nsegments:
            # skip empty
            if self.segments[idx].start == self.segments[idx].end: 
                idx += 1
                continue
            # no overlap - save segment
            # note the cange from > to >=
            if <PositionDifference>self.segments[idx].start - distance > max_end:
                self.segments[insertion_idx].end = max_end
                insertion_idx += 1
                self.segments[insertion_idx].start = self.segments[idx].start

            max_end = lmax( self.segments[idx].end, max_end )
            idx += 1

        # store final segment
        self.segments[insertion_idx].end = max_end

        # truncate array
        insertion_idx += 1
        self.nsegments = insertion_idx
        self._resize( self.nsegments )
        self.flag = 1

    cpdef check( self ):
        '''check if segment list is normalized.

        If it is, return True and set flag flag.
        '''

        # empty lists are normalized
        if self.nsegments == 0:
            self.flag = 1
            return self.flag

        self.flag = 0

        cdef int idx
        cdef Segment segment

        segment = self.segments[0]
        if segment.start >= segment.end:
            raise ValueError( "empty/invalid segment in segmentlist: %s" % str( segment ) )

        for idx from 1 <= idx < self.nsegments:
            segment = self.segments[idx]
            if segment.start >= segment.end:
                raise ValueError( "empty/invalid segment in segmentlist: %s" % str( segment ) )
            if self.segments[idx-1].start > segment.start:
                raise ValueError( "segment list is not sorted: %s > %s" % (self.segments[idx-1], segment) )
            if self.segments[idx-1].end > segment.start:
                raise ValueError( "segment overlap: %s overlaps %s" % (self.segments[idx-1], segment) )
            #if self.segments[idx-1].end == segment.start:
            #    raise ValueError( "segment touch: %s touches %s" % (self.segments[idx-1], segment) )

        self.flag = 1

        return self.flag

    cdef int _getInsertionPoint( self, Segment other ):
        '''return insertion point for other.

        The insertion point denotes the element at which
        or after which *other* should be inserted to keep
        the sort order.

        This method does not check if the segment list is normalized. To
        do this check, call getInsertionPoint().

        The function returns -1 if other is before the first element or the
        segment list is empty.
        The function returns nsegments if other is after the last element.
        '''
        cdef int idx
        if self.nsegments == 0: return -1

        # avoid out of range searches
        if other.start >= self.segments[self.nsegments-1].end:
            return self.nsegments
        if other.end <= self.segments[0].start: 
            return -1

        idx = searchsorted(self.segments,
                           self.nsegments,
                           sizeof(Segment),
                           &other,
                           &cmpSegments)

        if idx == self.nsegments:
            return idx-1
        elif self.segments[idx].start != other.start:
            return idx-1
        else:
            return idx

    def getInsertionPoint( self, start, end ):
        assert self.flag & SEG_NORMALIZED, "searching in non-normalized list"
        return self._getInsertionPoint( Segment( start, end ) )

    property isNormalized:
        def __get__(self): return self.flag & SEG_NORMALIZED

    property statusflag:
        def __get__(self): return self.flag

    property isEmpty:
        def __get__(self): return self.nsegments == 0

    cpdef Position getRandomPosition(self):
        '''return a random position within the workspace.

        Not efficient, see the specialized samplers instead if
        you want to sample repeatedly from the same workspace.
        '''
        cdef Position pos = numpy.random.randint( 0, self.sum() )
        cdef idx
        cdef Position l
        for idx from 0 <= idx < self.nsegments:
            l = self.segments[idx].end - self.segments[idx].start
            if pos > l:
                pos -= l
            else:
                return self.segments[idx].start + pos
        assert False

    cdef Position overlap(self, Segment other):
        """return the number of nucleotides overlapping with other.

        Arguments
        ---------
        other : Segment
            Segment to compute overlap with

        Returns
        ------
        overlap : int
            Number of residues overlapping
        """
        assert self.flag & SEG_NORMALIZED, "non-normalized segment list"

        cdef int idx
        idx = self._getInsertionPoint(other)

        # deal with border cases
        # permit partial overlap
        if idx == self.nsegments:
            idx -=1
        elif idx == -1:
            idx=0

        cdef Position count
        count = 0

        while idx < self.nsegments and self.segments[idx].start <= other.end:
            count += segment_overlap(self.segments[idx], other)
            idx += 1
        return count

    cdef SegmentList getOverlappingSegments(self, Segment other):
        """return the segments overlapping with other.

        Arguments
        ---------
        other : Segment
            Segment to compare to.

        Returns
        -------
        segmentlist : SegmentList
            A new segment list
        """

        assert self.flag & SEG_NORMALIZED, "non-normalized segment list"
        cdef int idx
        idx = self._getInsertionPoint(other)

        # deal with border cases
        # permit partial overlap
        if idx == self.nsegments:
            idx -=1
        elif idx == -1:
            idx=0

        count = 0
        cdef SegmentList result = SegmentList()
        
        while idx < self.nsegments and self.segments[idx].start <= other.end:
            result._add(self.segments[idx])
            idx += 1
        return result

    cpdef Position overlapWithRange(self, Position start, Position end):
        """return the number of nucleotides overlapping with a range.

        Arguments
        ---------
        start : int
            Begin of range.
        end : int
            End of range.

        Returns
        ------
        overlap : int
            Number of residues overlapping
        """

        cdef Segment s
        s = Segment(start, end)
        return self.overlap(s)

    cpdef SegmentList getOverlappingSegmentsWithRange(
        self, Position start, Position end):
        """return the segments overlapping with a range.

        Arguments
        ---------
        start : int
            Begin of range.
        end : int
            End of range.

        Returns
        -------
        segmentlist : SegmentList
            A new segment list
        """

        cdef Segment s
        s = Segment(start, end)
        return self.getOverlappingSegments(s)

    cpdef Position overlapWithSegments(self, SegmentList other):
        '''return the number of nucleotides overlapping between this and *other*.

        Both segment lists need to be normalized.
        '''

        assert self.flag & SEG_NORMALIZED, "intersection from non-normalized list"
        assert other.flag & SEG_NORMALIZED, "intersection with non-normalized list"

        # avoid self-self comparison
        if other.segments == self.segments:
            return self.sum()

        cdef int this_idx, other_idx, working_idx, last_this_idx, last_other_idx
        this_idx = other_idx = 0
        last_this_idx = last_other_idx = -1
        cdef Segment this_segment = Segment(0,0)
        cdef Segment other_segment = Segment(0,0)

        cdef Position overlap
        overlap = 0

        while this_idx < self.nsegments and other_idx < other.nsegments:

            # print "this=", this_idx, self.nsegments, "other=", other_idx, other.nsegments
            if last_this_idx != this_idx:
                this_segment = self.segments[this_idx]
                last_this_idx = this_idx
            if last_other_idx != other_idx:
                other_segment = other.segments[other_idx]
                last_other_idx = other_idx

            # print this_segment, other_segment
            # skip segments in this not overlapping other
            if this_segment.end <= other_segment.start:
                this_idx += 1
            # skip segments in other not overlapping this
            elif other_segment.end <= this_segment.start:
                other_idx += 1
            else:
                # deal with overlap
                overlap += segment_overlap_raw(this_segment, other_segment)
                if this_segment.end < other_segment.end:
                    this_idx += 1
                elif other_segment.end < this_segment.end:
                    other_idx += 1
                else:
                    this_idx += 1
                    other_idx += 1

        return overlap

    cpdef Position intersectionWithSegments(
        self,
        SegmentList other, 
        mode="base"):
        """return number of segments overlapping with *other*.

        `mode` can be either ``base`` or ``midpoint``. With ``base``,
        an overlap is recorded if a single ``base`` in self overlaps a
        interval in other. With ``midpoint``, an overlap is counted
        if the midpoint of a segment overlaps any interval in other.

        Arguments
        ---------
        other : SegmentList
             SegmentList to overlap with
        mode : string
             Overlap mode.

        Returns
        -------
        overlap : int
            Number of overlapping segments
        """

        assert self.flag & SEG_NORMALIZED, "intersection from non-normalized list"
        assert other.flag & SEG_NORMALIZED, "intersection with non-normalized list"

        # avoid self-self comparison
        if other.segments == self.segments: return self.sum()

        cdef int this_idx, other_idx, last_this_idx, last_other_idx
        this_idx = other_idx = 0
        last_this_idx = last_other_idx = -1
        cdef Segment this_segment = Segment(0,0)
        cdef Segment other_segment = Segment(0,0)

        cdef int midpoint_overlap = mode == "midpoint"

        cdef Position noverlap
        noverlap = 0

        while this_idx < self.nsegments and other_idx < other.nsegments:

            # print "this=", this_idx, self.nsegments, "other=", other_idx, other.nsegments
            if last_this_idx != this_idx:
                this_segment = self.segments[this_idx]
                last_this_idx = this_idx
            if last_other_idx != other_idx:
                other_segment = other.segments[other_idx]
                last_other_idx = other_idx

            # skip segments in this not overlapping other
            if this_segment.end <= other_segment.start:
                this_idx += 1
            # skip segments in other not overlapping this
            elif other_segment.end <= this_segment.start:
                other_idx += 1
            else:
                # deal with overlap
                if midpoint_overlap:
                    if other_segment.start <= this_segment.start + \
                       (this_segment.end - this_segment.start) // 2 \
                            < other_segment.end:
                        noverlap += 1
                else:
                    noverlap += 1
                this_idx += 1

        return noverlap

    def getLengthDistribution( self, bucket_size = 0, nbuckets = 100000):
        '''build histogram of segments lengths.
        
        If bucket_size is 0, the bucket size will be determined automatically
        to accomodate the largest segment.

        returns a histogram with the bucket_size.
        '''
        assert bucket_size >= 0, 'bucket_size is 0'
        assert nbuckets > 0, 'nbuckets is 0'

        cdef int idx, i
        cdef Position l
        cdef Segment s
        cdef numpy.ndarray[DTYPE_INT_t, ndim=1] histogram
        histogram = numpy.zeros( nbuckets, dtype=numpy.int )
        if bucket_size == 0:
            bucket_size = int(math.ceil(segment_length(self.largest()) / float(nbuckets)))
            
        for idx from 0 <= idx < self.nsegments:
            l = segment_length(self.segments[idx])
            i = <int>((l+bucket_size-1)/bucket_size)
            if i >= nbuckets:
                raise ValueError(
                    "segment %i-%i too large: %i >= %i, "
                    "increase nbuckets (%i) or bucket_size (%i) "
                    "such that nbuckets * bucket_size > %i" %
                    (self.segments[idx].start,
                      self.segments[idx].end,
                      l, 
                      nbuckets * bucket_size,
                      nbuckets,
                      bucket_size,
                      l
                  ))
            histogram[i] += 1
        return histogram, bucket_size

    cdef void truncate(self, Segment other):
        '''truncate Segment list to range given by segment *other*.'''
        
        cdef int idx
        cdef Segment * s
        cdef Position start = other.start
        cdef Position end = other.end

        for idx from 0 <= idx < self.nsegments:
            s = & self.segments[idx] 
            if s.end < start: s.start = s.end = 0
            elif s.start > end: s.start = s.end = 0
            else:
                if s.start < start: s.start = start
                if s.end > end: s.end = end

        self.normalize()

    cpdef void subtract(self, SegmentList other):
        '''subtract residues in *other* segment list with from this on.

        Intervals are truncated, any residues overlapping with *other*
        are removed.

        Both this and other are assumed to have been normalized.

        '''
        assert self.flag & SEG_NORMALIZED, "subtraction of a non-normalized list"
        assert other.flag & SEG_NORMALIZED, "subtraction with non-normalized list"

        # if self-self comparison, clear all residues
        if other.segments == self.segments: 
            self.clear()
            return

        if self.nsegments == 0:
            return

        cdef int this_idx, other_idx, working_idx, last_this_idx, last_other_idx
        working_idx = this_idx = other_idx = 0
        last_this_idx = last_other_idx = -1
        cdef Segment this_segment = Segment(0,0)
        cdef Segment other_segment = Segment(0,0)

        cdef Segment * new_segments
        cdef size_t allocated
        # create new list with 10% overhead
        allocated = int(lmax( self.nsegments, other.nsegments) * 1.1)

        new_segments =<Segment*>malloc( allocated * sizeof( Segment ) )
        if not new_segments: 
            raise MemoryError("out of memory when allocation %i bytes" %
                              sizeof(allocated * sizeof(Segment)))

        while this_idx < self.nsegments and other_idx < other.nsegments:

            # re-allocated before adding
            if working_idx >= allocated:
                allocated *= 2
                new_segments = <Segment*>realloc(new_segments, allocated * sizeof(Segment))
                assert new_segments != NULL

            # print "this=", this_idx, self.nsegments, this_segment, "other=", other_idx, other.nsegments, other.segments[other_idx]
            if last_this_idx != this_idx:
                this_segment = self.segments[this_idx]
                last_this_idx = this_idx
            if last_other_idx != other_idx:
                other_segment = other.segments[other_idx]
                last_other_idx = other_idx

            # print this_segment, other_segment
            # skip segments in this not overlapping other
            if this_segment.end <= other_segment.start:
                if this_segment.start < this_segment.end:
                    new_segments[working_idx] = this_segment
                    working_idx += 1
                this_idx += 1
            # skip segments in other not overlapping this
            elif other_segment.end <= this_segment.start:
                other_idx += 1
            else:
                # deal with overlap
                if this_segment.start < other_segment.start:
                    # print "adding", this_segment.start, other_segment.start
                    new_segments[working_idx].start = this_segment.start
                    new_segments[working_idx].end = other_segment.start
                    working_idx += 1

                this_segment.start = other_segment.end

        # output last
        if this_segment.start < this_segment.end:
            new_segments[working_idx].start = this_segment.start
            new_segments[working_idx].end = this_segment.end
            working_idx += 1

        free(self.segments)
        self.segments = new_segments
        self.nsegments = working_idx
        self.allocated = allocated

    cpdef summarize(self):
        '''returns summary statistics for segments.'''
        
        cdef int idx
        cdef Position min_length, max_length, length
        cdef Position total_length
        cdef Segment * s

        if self.nsegments == 0:
            return 0, 0, 0, 0

        # TODO: use maximum integer for Position
        min_length = UINT32_MAX
        max_length = 0
        total_length = 0

        for idx from 0 <= idx < self.nsegments:
            s = & self.segments[idx] 
            length = s.end - s.start
            total_length += length
            min_length = lmin( length, min_length )
            max_length = lmax( length, max_length )

        cdef double mean_length = <double>total_length / self.nsegments
                           
        return mean_length, min_length, max_length, total_length

    cpdef SegmentList getFilledSegmentsFromStart( self, Position start, PositionDifference remainder ):
        '''start filling segment from *start* until *remainder*
           bases have been covered. 

           This method wraps around at the ends if remainder can not
           be filled. If *remainder* is larger than the segment list,
           the segment list itself will be returned.

           return a segment list 
           '''

        if remainder > self.sum():
            return SegmentList(clone=self)

        cdef SegmentList result = SegmentList()

        cdef int idx
        idx = self.getInsertionPoint( start, start + 1 )
        cdef Position end
        
        # deal with border cases
        # permit partial overlap
        if idx == self.nsegments: idx -=1
        elif idx == -1: idx=0

        # add from start until complete
        while remainder > 0:
            if self.segments[idx].end < start: 
                pass
            else:
                start = lmax(self.segments[idx].start, start)
                end = lmin(self.segments[idx].end, start + remainder)
                remainder -= end - start
                result._add(Segment(start, end))
                
            idx += 1
            if idx == self.nsegments:
                idx = 0
                start = self.segments[idx].start

        result.normalize()
        return result

    cpdef SegmentList getFilledSegmentsFromEnd(
        self, Position end, PositionDifference remainder):
        '''start filling segment from *end* until *remainder*
           bases have been covered. Fill in reverse order.

           This method wraps around at the ends if remainder can not
           be filled. If *remainder* is larger than the segment list,
           the segment list itself will be returned.

           return a segment list 
           '''

        if remainder > self.sum():
            return SegmentList(clone=self)

        cdef SegmentList result = SegmentList()

        cdef int idx
        idx = self.getInsertionPoint( end, end + 1 )
        cdef Position start
        
        # deal with border cases
        # permit partial overlap
        if idx == self.nsegments: idx -=1
        elif idx == -1: idx=0

        # add from start until complete
        while remainder > 0:
            if self.segments[idx].start > end: 
                pass
            else:
                end = lmin( self.segments[idx].end, end )
                start = lmax( self.segments[idx].start, end - remainder )
                remainder -= end - start
                result._add( Segment( start, end ) )
                
            idx -= 1
            if idx < 0:
                idx = self.nsegments - 1
                end = self.segments[idx].end

        result.normalize()
        return result

    cpdef void filter(self, SegmentList other):
        '''remove all segments that are not in *other*

        Both this and other are assumed to have been normalized.
        '''

        # if self-self comparison, remove all
        if other.segments == self.segments:
            self.clear()
            return
        if self.nsegments == 0:
            return

        cdef Segment * new_segments
        cdef size_t allocated
        # create new list
        new_segments =<Segment*>malloc( self.nsegments * sizeof(Segment))
        if not new_segments: 
            raise MemoryError(
                "out of memory when allocation %i bytes" %
                sizeof(self.nsegments * sizeof(Segment)))

        cdef int this_idx, other_idx, working_idx, last_this_idx, last_other_idx
        working_idx = this_idx = other_idx = 0
        last_this_idx = last_other_idx = -1
        cdef Segment this_segment = Segment(0,0)
        cdef Segment other_segment = Segment(0,0)
        # for negative segments, do not use -1
        cdef Position last_start = self.segments[0].start - 1

        while this_idx < self.nsegments and other_idx < other.nsegments:

            # print "this=", this_idx, self.nsegments, "other=", other_idx, other.nsegments
            if last_this_idx != this_idx:
                this_segment = self.segments[this_idx]
                last_this_idx = this_idx
            if last_other_idx != other_idx:
                other_segment = other.segments[other_idx]
                last_other_idx = other_idx

            # print this_segment, other_segment
            # skip segments in this not overlapping other
            if this_segment.end <= other_segment.start:
                this_idx += 1
            # skip segments in other not overlapping this
            elif other_segment.end <= this_segment.start:
                other_idx += 1
            else:
                # deal with overlap
                if last_start != this_segment.start:
                    new_segments[working_idx].start = this_segment.start
                    new_segments[working_idx].end = this_segment.end
                    working_idx += 1
                    last_start = this_segment.start

                if this_segment.end < other_segment.end:
                    this_idx += 1
                elif other_segment.end < this_segment.end:
                    other_idx += 1
                else:
                    this_idx += 1
                    other_idx += 1

        free(self.segments)
        self.nsegments = working_idx
        self.segments = new_segments
        self._resize(self.nsegments)

    cpdef void intersect(self, SegmentList other):
        '''intersect this segment list with another.

        Intervals are truncated to the overlap between this and
        other.

        Both this and other are assumed to have been normalized.

        Segments are truncated to the smallest intersection. Thus,
        the intersection between ``[(0,10), (10,20)]`` and ``[(0,20)]``
        will be ``[(0,10), (10,20)]`` and not ``[(0,20)]``.

        '''
        assert self.flag & SEG_NORMALIZED, "intersection of a non-normalized list"
        assert other.flag & SEG_NORMALIZED, "intersection with non-normalized list"

        # avoid self-self comparison
        if other.segments == self.segments:
            return

        if self.nsegments == 0:
            return

        cdef Segment * new_segments
        cdef size_t allocated
        # create new list with 10% overhead
        allocated = int(lmax(self.nsegments, other.nsegments) * 1.1)

        new_segments =<Segment*>malloc(allocated * sizeof(Segment))
        if not new_segments: 
            raise MemoryError("out of memory when allocation %i bytes" %
                              sizeof(allocated * sizeof(Segment)))

        cdef int this_idx, other_idx, working_idx, last_this_idx, last_other_idx
        working_idx = this_idx = other_idx = 0
        last_this_idx = last_other_idx = -1
        cdef Segment this_segment = Segment(0,0)
        cdef Segment other_segment = Segment(0,0)

        while this_idx < self.nsegments and other_idx < other.nsegments:

            # print "this=", this_idx, self.nsegments, "other=", other_idx, other.nsegments
            if last_this_idx != this_idx:
                this_segment = self.segments[this_idx]
                last_this_idx = this_idx
            if last_other_idx != other_idx:
                other_segment = other.segments[other_idx]
                last_other_idx = other_idx

            # print this_segment, other_segment
            # skip segments in this not overlapping other
            if this_segment.end <= other_segment.start:
                this_idx += 1
            # skip segments in other not overlapping this
            elif other_segment.end <= this_segment.start:
                other_idx += 1
            else:
                # deal with overlap
                new_segments[working_idx].start = lmax(this_segment.start,
                                                       other_segment.start)
                new_segments[working_idx].end = lmin(this_segment.end,
                                                     other_segment.end)
                working_idx += 1
                if working_idx >= allocated:
                    allocated *= 2
                    new_segments = <Segment*>realloc(new_segments,
                                                     allocated * sizeof(Segment ) )
                    assert new_segments != NULL

                if this_segment.end < other_segment.end:
                    this_idx += 1
                elif other_segment.end < this_segment.end:
                    other_idx += 1
                else:
                    this_idx += 1
                    other_idx += 1

        free(self.segments)
        self.segments = new_segments
        self.nsegments = working_idx
        self.allocated = allocated

    cpdef extend_segments(self, int extension):
        '''extend all intervals by a certain amount.

        If the extension causes an interval to extend beyond the chromosome
        start, the interval is truncated.

        The segment list will not be normalized afterwards.'''

        cdef int idx
        cdef Segment * s
        for idx from 0 <= idx < self.nsegments:
            s = &self.segments[idx]
            s.start -= lmin( extension, s.start)
            s.end += extension
        self.flag = False

    cpdef expand_segments( self, double expansion ):
        '''expand all intervals by a certain amount.

        The interval length is multiplied by *expansion* and 
        extended around the mid-point.

        If the expansion causes an interval to extend beyond the chromosome
        start, the interval is truncated.

        The segment list will not be normalized afterwards.'''

        cdef int idx, 
        cdef PositionDifference l, extension
        cdef Segment * s
        cdef double e_1 = (expansion - 1.0) / 2.0
        
        if expansion <= 0: raise ValueError( "invalid expansion: %f <= 0" % expansion)

        for idx from 0 <= idx < self.nsegments:
            s = &self.segments[idx]
            l = s.end - s.start 
            extension = <PositionDifference>floor(l * e_1 )
            s.start -= lmin( extension, s.start)
            s.end += extension
        self.flag = False

    def shift(self, PositionDifference offset):
        '''shift segments by a certain offset.
        
        raises ValueError if segment coordinates become negative.
        '''
        cdef int idx
        cdef Segment * s
        for idx from 0 <= idx < self.nsegments:
            s = &self.segments[idx]
            s.end += offset
            s.start += offset
            if s.start < 0:
                raise ValueError("shift creates negative coordinates")

    cpdef Position sum(self):
        '''return total length of all segments.'''
        cdef Position total
        cdef int idx
        cdef Segment s
        total = 0
        for idx from 0 <= idx < self.nsegments:
            s = self.segments[idx]
            total += s.end - s.start
        return total

    def largest(self):
        '''return largest segment.

        If there are multiple segments of the same size, the first
        will be returned.
        '''
        if self.nsegments == 0:
            raise ValueError("largest segment from empty list")
        cdef int idx
        cdef int max_idx = 0
        cdef Position max_value = 0
        cdef Position l
        total = 0
        for idx from 0 <= idx < self.nsegments:
            l = segment_length( self.segments[idx] )
            if l > max_value:
                max_idx = idx
                max_value = l
        return self.segments[max_idx]

    cpdef Position max(self):
        '''return maximum coordinate.'''
        assert self.flag & SEG_NORMALIZED, "maximum from non-normalized list"
        if self.nsegments == 0: return 0
        return self.segments[self.nsegments - 1].end

    cpdef Position min(self):
        '''return minimum coordinate.'''
        assert self.flag & SEG_NORMALIZED, "minimum from non-normalized list"
        if self.nsegments == 0: return 0
        return self.segments[0].start

    cpdef clear(self):
        '''removes all segments from list.

        Note that this method does not shrink the allocated memory.
        '''
        self.nsegments = 0
        self.flag = 1

    cpdef SegmentList clone(self):
        """return a copy of self."""
        clone = SegmentList(allocate=self.nsegments)
        memcpy(clone.segments,
               self.segments,
               self.nsegments * sizeof(Segment))
        clone.nsegments = self.nsegments
        clone.flag = self.flag
        return clone

    def __len__(self):
        return self.nsegments

    def __dealloc__(self):
        cdef int fd
        
        if self.segments != NULL:
            # unshared memory
            if self.is_shared:
                if self.is_slave:
                    pass
                else:
                    # free shared memory as master
                    munmap(self.segments, 
                           self.nsegments * sizeof(Segment))
            else:
                # free private copy
                free(self.segments)

        if self.shared_fd != -1:
            fd = shm_unlink(self.key)
            if fd == -1:
                raise OSError("could not unlink shared memory")

    def __str__(self):
        return str(self.asList())

    def asList( self ):

        cdef int idx
        result = []
        for idx from 0 <= idx < self.nsegments:
            result.append((self.segments[idx].start, self.segments[idx].end))
        return result

    def asLengths( self ):
        cdef int idx
        result = []
        for idx from 0 <= idx < self.nsegments:
            result.append( self.segments[idx].end - self.segments[idx].start)
        return result

    def __iter__( self ):
        return SegmentListIterator(self)

    def __getitem__(self, key):
        if key >= self.nsegments:
            raise IndexError("index out of range")
        return self.segments[key].start, self.segments[key].end
    
    def compare(self, SegmentList other):
        cdef int idx
        cdef int l1 = self.__len__()
        if other is None:
            return -1
        cdef int l2 = len(other)
        if l2 - l1 != 0:
            return l2 - l1
        for idx from 0 <= idx < self.nsegments:
            x = cmpSegmentsStartAndEnd(&self.segments[idx], &other.segments[idx])
            if x != 0:
                return x
        return 0

    def __richcmp__(self, SegmentList other, int op):
        if op == 2:  # == operator
            return self.compare(other) == 0
        elif op == 3:  # != operator
            return self.compare(other) != 0
        else:
            return NotImplemented


def buildSegmentList(*args):
    '''pickling helper function.
    
    Return a re-constructed SegmentList object.
    '''
    return SegmentList(unreduce=args)

# SegmentTuple = collections.namedtuple( "SegmentTuple", "start, end" )

cdef class SegmentListIterator:

    cdef SegmentList segment_list
    cdef Position idx

    def __init__(self, SegmentList container):
        self.idx = 0
        self.segment_list = container

    def __iter__(self):
        return self

    def __next__(self):
        cdef Position t
        cdef Segment v
        if self.idx >= self.segment_list.nsegments:
            raise StopIteration
        v = self.segment_list.segments[self.idx]
        self.idx += 1
        return v.start, v.end




'''
IOTools - tools for I/O operations
==================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''

import string
import re
import sys
import os
import collections
import types
import glob
import stat
import gzip

import numpy
import numpy.ma

########################################################################


def readMap(infile,
            columns=(0, 1),
            map_functions = (str, str),
            both_directions=False,
            has_header = False):
    """read a map (pairs of values) from infile.
    returns a hash.

    Use map functions to convert elements.
    If both_directions is set to true, both mapping directions are returned.
    """
    m = {}
    r = {}
    n = 0

    if columns == "all":
        key_column = 0
        value_column = None
    else:
        key_column, value_column = columns

    key_function, value_function = map_functions

    for l in infile:
        if l[0] == "#":
            continue
        n += 1

        if has_header and n == 1:
            continue

        d = l[:-1].split("\t")
        if len(d) < 2:
            continue
        key = key_function(d[key_column])
        if value_column:
            val = value_function(d[value_column])
        else:
            val = tuple(map(value_function, [d[x] for x in range(1, len(d))]))

        m[key] = val
        if val not in r:
            r[val] = []
        r[val].append(key)

    if both_directions:
        return m, r
    else:
        return m

########################################################################


def readList(infile,
             column=0,
             map_function=str,
             map_category={},
             with_title=False):
    """read a list of values from infile.

    Use map_function to convert values.
    Use map_category, to map read values directory 
    If with_title, first line is assumed to be a title
    """

    m = []
    errors = []
    title = None
    for l in infile:
        if l[0] == "#":
            continue
        if with_title and not title:
            title = l[:-1].split("\t")[column]
            continue

        try:
            d = map_function(l[:-1].split("\t")[column])
        except ValueError:
            continue

        if map_category:
            d = map_category[d]
        m.append(d)

    if with_title:
        return title, m
    else:
        return m

########################################################################


def ReadList(infile, column=0, map_function=str, map_category={}):
    """read a list of values from infile.

    Use map_function to convert values.
    Use map_category, to map read values directory 
    """

    m = []
    errors = []
    for l in infile:
        if l[0] == "#":
            continue
        try:
            d = map_function(l[:-1].split("\t")[column])
        except ValueError:
            continue
        try:
            if map_category:
                d = map_category[d]

        except KeyError:
            errors.append(d)
            continue
        m.append(d)

    return m, errors

########################################################################


def readMultiMap(infile,
                 columns=(0, 1),
                 map_functions = (str, str),
                 both_directions=False,
                 has_header = False,
                 dtype = dict):
    """read a map (pairs of values) from infile.
    returns a hash. 

    Use map functions to convert elements.
    If both_directions is set to true, both mapping directions are returned.
    This function can have n:n matches
    """
    m = dtype()
    r = dtype()
    n = 0
    for l in infile:
        if l[0] == "#":
            continue
        n += 1

        if has_header and n == 1:
            continue

        d = l[:-1].split("\t")
        try:
            key = map_functions[0](d[columns[0]])
            val = map_functions[1](d[columns[1]])
        except (ValueError, IndexError) as msg:
            raise ValueError("parsing error in line %s: %s" % (l[:-1], msg))

        if key not in m:
            m[key] = []
        m[key].append(val)
        if val not in r:
            r[val] = []
        r[val].append(key)

    if both_directions:
        return m, r
    else:
        return m

########################################################################
# Read a table


def readTable(file,
              separator="\t",
              numeric_type=numpy.float,
              take="all",
              headers=True,
              truncate=None,
              cumulate_out_of_range=True,
              ):
    """read a matrix. There probably is a routine for this in Numpy, which
    I haven't found yet.

    If cumulate_out_of_range is set to true, the terminal bins will
    contain the cumulative values of bins out of range.

    """

    lines = [x for x in file.readlines() if x[0] != "#"]

    if len(lines) == 0:
        return None, []

    if take == "all":
        num_cols = len(string.split(lines[0][:-1], "\t"))
        take = list(range(0, num_cols))
    else:
        num_cols = len(take)

    if headers:
        headers = lines[0][:-1].split("\t")
        headers = [headers[x] for x in take]
        del lines[0]

    num_rows = len(lines)
    matrix = numpy.ma.masked_array(
        numpy.zeros((num_rows, num_cols), numeric_type))

    if truncate:
        min_row, max_row = truncate

    nrow = 0
    min_data = [0] * num_cols
    max_data = None
    for l in lines:
        data = l[:-1].split("\t")
        data = [data[x] for x in take]

        # try conversion. Unparseable fields set to missing_value
        for x in range(len(data)):
            try:
                data[x] = float(data[x])
            except ValueError:
                data[x] = numpy.ma.masked

        if truncate != None:
            if data[0] < min_row:
                if cumulate_out_of_range:
                    for x in range(1, num_cols):
                        min_data[x] += data[x]
                continue
            elif data[0] >= max_row:
                if max_data == None:
                    max_data = [0] * num_cols
                    max_data[0] = max_row
                for x in range(1, num_cols):
                    try:
                        max_data[x] += data[x]
                    except TypeError:
                        # missing values cause type errors
                        continue
                continue
            elif min_row != None:
                if cumulate_out_of_range:
                    for x in range(0, num_cols):
                        try:
                            min_data[x] += data[x]
                        except TypeError:
                            # missing values cause type errors
                            continue
                    else:
                        min_data = data
                data = min_data
                min_row = None

        # copy values into matrix
        # this is a bit clumsy, but missing values
        # cause an error otherwise
        for x in range(len(data)):
            matrix[nrow, x] = data[x]

        nrow += 1

    if truncate != None:
        if cumulate_out_of_range:
            if max_data != None:
                matrix[nrow] = max_data

        # truncate matrix
        matrix = matrix[0:nrow + 1, 0:num_cols]

    return matrix, headers

########################################################################


def getInvertedDictionary(dict, make_unique=False):
    """returns an inverted dictionary with keys and values swapped.
    """
    inv = {}
    if make_unique:
        for k, v in dict.items():
            inv[v] = k
    else:
        for k, v in dict.items():
            inv.setdefault(v, []).append(k)
    return inv

########################################################################
# Read a sequence from a fasta file


def readSequence(file):
    """read sequence from a fasta file.

    returns a tuple with description and sequence
    """

    s = []
    for line in file:
        if line[0] == ">":
            description = line[1:-1]
            continue
        elif line[0] == "#":
            continue
        else:
            s.append(re.sub("\s", "", line[:-1]))

    return description, "".join(s)


def getLastLine(filename, read_size=1024):
    """return last line of a file.
    """

    # U is to open it with Universal newline support
    f = open(filename, 'rU')
    offset = read_size
    f.seek(0, 2)
    file_size = f.tell()
    if file_size == 0:
        return ""
    while 1:
        if file_size < offset:
            offset = file_size
        f.seek(-1 * offset, 2)
        read_str = f.read(offset)
        # Remove newline at the end
        if read_str[offset - 1] == '\n':
            read_str = read_str[:-1]
        lines = read_str.split('\n')
        if len(lines) >= 2:
            return lines[-1]
        if offset == file_size:   # reached the beginning
            return read_str
        offset += read_size
    f.close()


def ReadMap(*args, **kwargs):
    """compatibility - see readMap."""
    return readMap(*args, **kwargs)


def isEmpty(filename):
    '''return True if file exists and is empty.

    raises OSError if file does not exist
    '''
    return os.stat(filename)[stat.ST_SIZE] == 0


def prettyFloat(val, format="%5.2f"):
    """output a float or "na" if not defined"""
    try:
        x = format % val
    except (ValueError, TypeError):
        x = "na"
    return x


def prettyPercent(numerator, denominator, format="%5.2f"):
    """output a percent value or "na" if not defined"""
    try:
        x = format % (100.0 * numerator / denominator)
    except (ValueError, ZeroDivisionError):
        x = "na"
    return x


def prettyString(val):
    '''output val or na if val == None'''
    if val != None:
        return val
    else:
        return "na"


class nested_dict(collections.defaultdict):

    """
    Auto-vivifying nested dict
    E.g.::
      nd= nested_dict()
      nd["mouse"]["chr1"]["+"] = 311 
    """

    def __init__(self):
        collections.defaultdict.__init__(self, nested_dict)

    def iterflattened(self):
        """
        iterate through values with nested keys flattened into a tuple
        """

        for key, value in self.items():
            if isinstance(value, nested_dict):
                for keykey, value in value.iterflattened():
                    yield (key,) + keykey, value
            else:
                yield (key,), value


def flatten(l, ltypes=(list, tuple)):
    ltype = type(l)
    l = list(l)
    i = 0
    while i < len(l):
        while isinstance(l[i], ltypes):
            if not l[i]:
                l.pop(i)
                i -= 1
                break
            else:
                l[i:i + 1] = l[i]
        i += 1
    return ltype(l)


def which(program):
    """check if program is in path.

    from post at http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
    """

    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def convertValue(value):
    '''convert a value to int, float or str.'''
    rx_int = re.compile("^\s*[+-]*[0-9]+\s*$")
    rx_float = re.compile("^\s*[+-]*[0-9.]+[.+\-eE][+-]*[0-9.]*\s*$")

    if value == None:
        return value

    if "," in value:
        values = []
        for value in value.split(","):
            if rx_int.match(value):
                values.append(int(value))
            elif rx_float.match(value):
                values.append(float(value))
            else:
                values.append(value)
        return values
    else:
        if rx_int.match(value):
            return int(value)
        elif rx_float.match(value):
            return float(value)
        return value


def iterate_tabular(infile, sep="\t"):
    '''iterate over infile skipping comments.'''
    for line in infile:
        if line.startswith("#"):
            continue
        yield line[:-1].split(sep)


def openFile(filename, mode="r", create_dir=False, encoding="utf-8"):
    '''open file called *filename* with mode *mode*.

    gzip - compressed files are recognized by the
    suffix ``.gz`` and opened transparently.

    Note that there are differences in the file
    like objects returned, for example in the
    ability to seek.

    Arguments
    ---------
    filename : string
    mode : string
       File opening mode
    create_dir : bool
       If True, the directory containing filename
       will be created if it does not exist.

    Returns
    -------
    File or file-like object in case of gzip compressed files.
    '''

    _, ext = os.path.splitext(filename)

    if create_dir:
        dirname = os.path.dirname(filename)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

    if ext.lower() in (".gz", ".z"):
        if sys.version_info.major >= 3:
            if mode == "r":
                return gzip.open(filename, 'rt', encoding=encoding)
            elif mode == "w":
                return gzip.open(filename, 'wt', encoding=encoding)
            elif mode == "a":
                return gzip.open(filename, 'wt', encoding=encoding)
        else:
            return gzip.open(filename, mode)
    else:
        if sys.version_info.major >= 3:
            return open(filename, mode, encoding=encoding)
        else:
            return open(filename, mode)




# cython: embedsignature=True
# cython: profile=False

cimport cython

from posix.types cimport off_t

cdef class CoordinateList:
    
    def __init__(self):
        self.is_shared = False
        self.is_slave = False

    cdef void fromMMAP(self):
        '''retrieve data from mmapped location to private copy.
        '''
        raise NotImplementedError()

    cdef off_t toMMAP(self, 
                      void * mmap, 
                      int shared_fd, 
                      off_t offset):
        '''move contents of list to shared memory 
        at *mmap* with *offset*.
        '''
        raise NotImplementedError()



from posix.types cimport off_t

cdef class CoordinateList:

    # Flag indicating if data is residing in shared memory
    cdef bint is_shared
    # Flag if object is owner of data or not.
    cdef bint is_slave
    # Chunk size for allocating an empty object
    cdef int chunk_size
    # File descriptor of shared memory
    cdef int shared_fd
    # Key to identify shared memory segment
    cdef key

    cdef off_t toMMAP(self, void *, int, off_t)
    cdef void fromMMAP(self)



#!/usr/bin/env python
##########################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##########################################################################
'''
great - run great analysis
==========================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script compares one or more genomic segments of interest 
against one more other genomic annotations.

GREAT says that a segment overlaps an annotation if the midpoint
of the segment overlaps the annotation.

The GREAT analysis is done on a per-isochore basis and overall.
For the overall analysis, there is no isochore correction
as the union of all isochores is treated as a single workspace.

This script implies several counters (``--counter`` option):

binom
    binomial model (as implemented by GREAT). Returns the probability
    of a certain number of segments overlapping (mid-point overlap) 
    an annotation

hyperg
    hypergeometric model. Returns the probabibility of a certain
    number of nucleotides within segments overlapping an annotation.
    Conceptually, all genomic positions are treated independently and 
    segments are fragmented into a collection of single bases that are
    placed indepently. This is valid for point intervals such as SNPs,
    but of course ignores segment size, though it might be used as a
    guide towards the expected fold enrichment, that can be computed
    more accurately using :mod:`gat-run.py`

Usage
-----

Example::

   python gat-great.py 
      --segment-file=segments.bed.gz 
      --workspace-file=workspace.bed.gz 
      --annotation-file=annotations_architecture.bed.gz  
 
Type::

   python gat-run.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os
import sys
import re
import optparse
import collections
import types
import glob
import time
import numpy
import numpy.random

import gat
import gat.Experiment as E
import gat.IOTools as IOTools
import gat.IO as IO
import GatSegmentList

import scipy.stats

GREAT_RESULT = collections.namedtuple("GREAT",
                                      ("track",
                                       "annotation",
                                       "isochore",
                                       "counter",
                                       # nsegments_overlapping_annotations
                                       "observed",
                                       "expected",
                                       "nsegments_in_workspace",
                                       "nannotations_in_workspace",
                                       "nsegments_overlapping_annotation",
                                       "nannotations_overlapping_segments",
                                       "basecoverage_intersection",
                                       "basecoverage_segments",
                                       "basecoverage_annotation",
                                       "basecoverage_workspace",
                                       "fraction_coverage_annotation",
                                       "fold",
                                       "pvalue",
                                       "qvalue"))


def main(argv):

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = optparse.OptionParser(version="%prog version: $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $",
                                   usage=globals()["__doc__"])

    parser.add_option("-a", "--annotation-file", "--annotations", dest="annotation_files", type="string", action="append",
                      help="filename with annotations [default=%default].")

    parser.add_option("-s", "--segment-file", "--segments", dest="segment_files", type="string", action="append",
                      help="filename with segments. Also accepts a glob in parentheses [default=%default].")

    parser.add_option("-w", "--workspace-file", "--workspace", dest="workspace_files", type="string", action="append",
                      help="filename with workspace segments. Also accepts a glob in parentheses [default=%default].")

    parser.add_option("-i", "--isochore-file", "--isochores", dest="isochore_files", type="string", action="append",
                      help="filename with isochore segments. Also accepts a glob in parentheses [default=%default].")

    parser.add_option("-o", "--order", dest="output_order", type="choice",
                      choices=(
                          "track", "annotation", "fold", "pvalue", "qvalue"),
                      help="order results in output by fold, track, etc. [default=%default].")

    parser.add_option("-q", "--qvalue-method", dest="qvalue_method", type="choice",
                      choices=(
                          "storey", "BH", "bonferroni", "holm", "hommel", "hochberg", "BY", "none"),
                      help="method to perform multiple testing correction by controlling the fdr [default=%default].")

    parser.add_option("--qvalue-lambda", dest="qvalue_lambda", type="float",
                      help="fdr computation: lambda [default=%default].")

    parser.add_option("--qvalue-pi0-method", dest="qvalue_pi0_method", type="choice",
                      choices=("smoother", "bootstrap"),
                      help="fdr computation: method for estimating pi0 [default=%default].")
    parser.add_option("--descriptions", dest="input_filename_descriptions", type="string",
                      help="filename mapping annotation terms to descriptions. "
                      " if given, the output table will contain additional columns "
                      " [default=%default]")

    parser.add_option("--ignore-segment-tracks", dest="ignore_segment_tracks", action="store_true",
                      help="ignore segment tracks - all segments belong to one track [default=%default]")

    parser.add_option("--enable-split-tracks", dest="enable_split_tracks", action="store_true",
                      help="permit the same track to be in multiple files [default=%default]")

    parser.add_option("--output-bed", dest="output_bed", type="choice", action="append",
                      choices=("all",
                               "annotations", "segments",
                               "workspaces", "isochores",
                               "overlap"),
                      help="output bed files [default=%default].")

    parser.add_option("--output-stats", dest="output_stats", type="choice", action="append",
                      choices=("all",
                               "annotations", "segments",
                               "workspaces", "isochores",
                               "overlap"),
                      help="output overlap summary stats [default=%default].")

    parser.add_option("--restrict-workspace", dest="restrict_workspace", action="store_true",
                      help="restrict workspace to those segments that contain both track"
                      " and annotations [default=%default]")

    parser.add_option("--counter", dest="counters", type="choice", action="append",
                      choices=("binom", "hyperg"),
                      help="counter to use [default=%default].")

    parser.add_option("--output-tables-pattern", dest="output_tables_pattern", type="string",
                      help="output pattern for result tables. Used if there are multiple counters used [default=%default].")

    parser.set_defaults(
        annotation_files=[],
        segment_files=[],
        workspace_files=[],
        sample_files=[],
        counters=[],
        output_stats=[],
        output_bed=[],
        output_tables_pattern="%s.tsv.gz",
        output_order="fold",
        input_filename_counts=None,
        input_filename_results=None,
        pvalue_method="empirical",
        output_plots_pattern=None,
        output_samples_pattern=None,
        qvalue_method="storey",
        qvalue_lambda=None,
        qvalue_pi0_method="smoother",
        ignore_segment_tracks=False,
        input_filename_descriptions=None,
        conditional="unconditional",
        conditional_extension=None,
        conditional_expansion=None,
        restrict_workspace=False,
        enable_split_tracks=False,
        shift_expansion=2.0,
        shift_extension=0,
        overlap_mode="midpoint",
        truncate_workspace_to_annotations=False,
        truncate_segments_to_workspace=False
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    tstart = time.time()

    if len(options.counters) == 0:
        options.counters.append("binom")

    ############################################
    segments, annotations, workspaces, isochores = IO.buildSegments(options)
    E.info("intervals loaded in %i seconds" % (time.time() - tstart))

    # filter segments by workspace
    workspace = IO.applyIsochores(
        segments, annotations, workspaces, options, isochores)

    ############################################
    description_header, descriptions, description_width = IO.readDescriptions(
        options)

    ############################################
    ############################################
    # compute per contig

    # compute bases covered by workspace
    workspace2basecoverage, isochores = {}, []
    for contig, ww in workspace.iteritems():
        workspace2basecoverage[contig] = ww.sum()
        isochores.append(contig)

    # compute percentage of bases covered by annotations in workspace
    # per isochore
    annotation2basecoverage = collections.defaultdict(dict)
    for annotation, aa in annotations.iteritems():
        for isochore, a in aa.iteritems():
            # need to truncate to workspace?
            annotation2basecoverage[annotation][isochore] = a.sum()

    results_per_contig = collections.defaultdict(list)

    E.info("computing counts per isochore")
    # results per isochore

    def emptyResult(segment, annotation, isochore,
                    counter,
                    nsegments_in_workspace,
                    basecoverage_annotation,
                    basecoverage_workspace):
        return GREAT_RESULT._make((
            segment, annotation, isochore,
            counter,
            0,  # observed
            0,  # expected
            nsegments_in_workspace,
            0,  # nannotations_in_workspace
            0,  # nsegments_overlapping_annotation
            0,  # nannotations_overlapping_segments
            0,  # basecoverage_intersection
            0,  # basecoverage_segments
            basecoverage_annotation,
            basecoverage_workspace,
            0.0,
            1.0,
            1.0,
            1.0))

    for isochore in isochores:
        basecoverage_workspace = workspace2basecoverage[isochore]

        # iterate over all isochores
        for segment, segmentdict in segments.iteritems():
            try:
                ss = segmentdict[isochore]
                # select segments overlapping workspace
                segments_in_workspace = GatSegmentList.SegmentList(clone=ss)
                segments_in_workspace.intersect(workspace[isochore])
                # number of segments in workspace
                nsegments_in_workspace = len(segments_in_workspace)
            except KeyError:
                ss = None

            basecoverage_segments = segments_in_workspace.sum()

            for annotation, annotationdict in annotations.iteritems():

                # if annotation != "GO:0030957": continue

                try:
                    aa = annotationdict[isochore]
                except KeyError:
                    aa = None

                # p_A: proportion of bases covered by annotation
                try:
                    basecoverage_annotation = annotation2basecoverage[
                        annotation][isochore]
                except KeyError:
                    basecoverage_annotation = 0

                if ss == None or aa == None:
                    for counter in options.counters:
                        results_per_contig[(counter, segment, annotation)].append(emptyResult(segment, annotation,
                                                                                              isochore,
                                                                                              counter,
                                                                                              nsegments_in_workspace,
                                                                                              basecoverage_annotation,
                                                                                              basecoverage_workspace))
                    continue

                # select segments overlapping annotation
                segments_overlapping_annotation = GatSegmentList.SegmentList(
                    clone=ss)
                segments_overlapping_annotation.intersect(
                    annotations[annotation][isochore])
                # number of segments in annotation
                nsegments_overlapping_annotation = ss.intersectionWithSegments(annotations[annotation][isochore],
                                                                               mode=options.overlap_mode)

                # number of nucleotides at the intersection of segments,
                # annotation and workspace
                basecoverage_intersection = segments_overlapping_annotation.sum()

                annotations_overlapping_segments = GatSegmentList.SegmentList(
                    clone=aa)
                annotations_overlapping_segments.intersect(ss)
                nannotations_overlapping_segments = len(
                    annotations_overlapping_segments)

                nannotations_in_workspace = len(aa)
                if nannotations_in_workspace == 0:
                    for counter in options.counters:
                        results_per_contig[(counter, segment, annotation)].append(emptyResult(segment,
                                                                                              annotation,
                                                                                              isochore,
                                                                                              counter,
                                                                                              nsegments_in_workspace,
                                                                                              basecoverage_annotation,
                                                                                              basecoverage_workspace))
                    continue

                fraction_coverage_annotation = basecoverage_annotation / \
                    float(basecoverage_workspace)
                fraction_hit_annotation = float(
                    nannotations_overlapping_segments) / nannotations_in_workspace

                for counter in options.counters:
                    if counter.startswith("binom"):
                        # GREAT binomial probability over "regions"
                        # n = number of genomic regions = nannotations_in_workspace
                        # ppi = fraction of genome annotated by annotation = fraction_coverage_annotation
                        # kpi = genomic regions with annotation hit by segments = nannotations_in_segments
                        # sf = survival functions = 1 -cdf
                        # probability of observing >kpi in a sample of n where the probabily of succes is
                        # ppi.
                        pvalue = scipy.stats.binom.sf(nsegments_overlapping_annotation - 1,
                                                      nsegments_in_workspace,
                                                      fraction_coverage_annotation)

                        expected = fraction_coverage_annotation * \
                            nsegments_in_workspace
                        observed = nsegments_overlapping_annotation

                    elif counter.startswith("hyperg"):

                        # hypergeometric probability over nucleotides
                        # Sampling without replacement
                        # x,M,n,M
                        # x = observed number of nucleotides in overlap of segments,annotations and workspace
                        # M = number of nucleotides in workspace
                        # n = number of nucleotides in annotations (and workspace)
                        # N = number of nucleotides in segments (and workspace)
                        # P-value of obtaining >x number of nucleotides
                        # overlapping.
                        rv = scipy.stats.hypergeom(basecoverage_workspace,
                                                   basecoverage_annotation,
                                                   basecoverage_segments)

                        pvalue = rv.sf(basecoverage_intersection)
                        expected = rv.mean()
                        observed = basecoverage_intersection

                    if expected != 0:
                        fold = float(observed) / expected
                    else:
                        fold = 1.0

                    r = GREAT_RESULT._make((
                        segment, annotation, isochore,
                        counter,
                        observed,
                        expected,
                        nsegments_in_workspace,
                        nannotations_in_workspace,
                        nsegments_overlapping_annotation,
                        nannotations_overlapping_segments,
                        basecoverage_intersection,
                        basecoverage_segments,
                        basecoverage_annotation,
                        basecoverage_workspace,
                        fraction_coverage_annotation,
                        fold,
                        pvalue,
                        1.0))
                    # print "\t".join( map(str, r))
                    results_per_contig[
                        (counter, segment, annotation)].append(r)

    E.info("merging counts per isochore")

    # compute sums
    results = []

    for niteration, pair in enumerate(results_per_contig.iteritems()):

        counter, segment, annotation = pair[0]
        data = pair[1]

        nsegments_in_workspace = sum([x.nsegments_in_workspace for x in data])
        nsegments_overlapping_annotation = sum([x.observed for x in data])
        nannotations_in_workspace = sum(
            [x.nannotations_in_workspace for x in data])
        nannotations_overlapping_segments = sum(
            [x.nannotations_overlapping_segments for x in data])

        basecoverage_intersection = sum(
            [x.basecoverage_intersection for x in data])
        basecoverage_segments = sum([x.basecoverage_segments for x in data])
        basecoverage_annotation = sum(
            [x.basecoverage_annotation for x in data])
        basecoverage_workspace = sum([x.basecoverage_workspace for x in data])

        fraction_coverage_annotation = basecoverage_annotation / \
            float(basecoverage_workspace)

        if counter.startswith("binom"):
            pvalue = scipy.stats.binom.sf(nsegments_overlapping_annotation - 1,
                                          nsegments_in_workspace,
                                          fraction_coverage_annotation)
            expected = fraction_coverage_annotation * nsegments_in_workspace
            observed = nsegments_overlapping_annotation
        elif counter.startswith("hyperg"):
            rv = scipy.stats.hypergeom(basecoverage_workspace,
                                       basecoverage_annotation,
                                       basecoverage_segments)

            pvalue = rv.sf(basecoverage_intersection)
            expected = rv.mean()
            observed = basecoverage_intersection

        if expected != 0:
            fold = float(observed) / expected
        else:
            fold = 1.0

        r = GREAT_RESULT._make((
            segment, annotation, "all",
            counter,
            observed,
            expected,
            nsegments_in_workspace,
            nannotations_in_workspace,
            nsegments_overlapping_annotation,
            nannotations_overlapping_segments,
            basecoverage_intersection,
            basecoverage_segments,
            basecoverage_annotation,
            basecoverage_workspace,
            fraction_coverage_annotation,
            fold,
            pvalue,
            1.0))

        results.append(r)

    IO.outputResults(results,
                     options,
                     GREAT_RESULT._fields,
                     description_header,
                     description_width,
                     descriptions)

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""test gat implementation."""

import unittest
import os
import numpy

from gat.Engine import AnnotatorResult, IntervalCollection, \
    Samples, SamplesCached, computeFDR, \
    SamplerAnnotator

from gat.SegmentList import SegmentList


class GatTest(unittest.TestCase):

    def shortDescription(self):
        return None


class TestIntervalCollection(GatTest):

    def setUp(self):
        self.a = IntervalCollection("a")
        self.a.add("track1", "contig1",
                   SegmentList(iter=((x, x + 10) for x in range(0, 1000, 100)), normalize=True))
        self.a.add("track1", "contig2",
                   SegmentList(iter=((x, x + 10) for x in range(0, 1000000, 100)), normalize=True))
        self.a.add("track2", "contig1",
                   SegmentList(iter=((x, x + 10) for x in range(1000, 2000, 100)), normalize=True))

    def testSaveLoad(self):

        fn = "tmp_testSaveLoad.bed"
        outfile = open(fn, "w")
        self.a.save(outfile)
        outfile.close()

        b = IntervalCollection("b")
        b.load(fn)

        self.assertEqual(len(self.a), len(b))
        self.assertEqual(self.a.tracks, b.tracks)
        self.assertEqual(self.a.sum(), b.sum())
        for t in b.tracks:
            self.assertEqual(sorted(self.a[t].keys()), sorted(b[t].keys()))
            for x in list(self.a[t].keys()):
                self.assertEqual(self.a[t][x], b[t][x])

    def testSharing(self):

        aa = self.a.clone()
        aa.share()


class TestToFromIsochores(GatTest):

    def setUp(self):

        a = IntervalCollection("a")

        # every ten at
        a.add("track1", "contig1",
              SegmentList(iter=((x, x + 250) for x in range(0, 2000, 500)), normalize=True))

        a.add("track1", "contig2",
              SegmentList(iter=((x, x + 250) for x in range(0, 2000, 500)), normalize=True))

        self.a = a

    def check(self, a, orig):

        self.assertEqual(a.tracks, orig.tracks)
        self.assertEqual(sorted(a["track1"].keys()),
                         sorted(['contig2', 'contig1']))

        for contig in ("contig1", "contig2"):
            self.assertEqual(list(a["track1"][contig]),
                             list(orig["track1"][contig]))

    def testNoIsochores(self):

        a = self.a.clone()
        orig = a.clone()
        a.fromIsochores()
        self.check(a, orig)

    def testToFromIsochores(self):

        a = self.a.clone()
        orig = a.clone()

        isochores = IntervalCollection("isochores")

        # covering isochores of size 100
        isochores.add("highGC", "contig1",
                      SegmentList(iter=((x, x + 100) for x in range(0, 10000, 200)), normalize=True))

        isochores.add("lowGC", "contig1",
                      SegmentList(iter=((x, x + 100) for x in range(100, 10000, 200)), normalize=True))

        isochores.add("highGC", "contig2",
                      SegmentList(iter=((x, x + 100) for x in range(0, 10000, 200)), normalize=True))

        isochores.add("lowGC", "contig2",
                      SegmentList(iter=((x, x + 100) for x in range(100, 10000, 200)), normalize=True))

        a.toIsochores(isochores)
        self.assertEqual(a.tracks, orig.tracks)
        self.assertEqual(sorted(a["track1"].keys()),
                         sorted(['contig2.highGC', 'contig1.highGC', 'contig2.lowGC', 'contig1.lowGC']))

        a.fromIsochores()

        self.check(a, orig)


class TestPValue(GatTest):
    '''test if pvalue computation is correct.'''

    def testPValue1(self):

        observed = 0.332640195285
        values = [0.3593727449353678, 0.24446041723385858, 0.11321078358680142, 0.28500665546177717, 0.017634423032620888, 0.47144573882791929, 0.20295266762886535, 0.24374906675401431, 0.12536987767373536, 0.36647407597049514, 0.1317950839045143, 0.32036858313479905, 0.2131875486832529, 0.18211958887292382, 0.4382662865088186, 0.12487068923091568, 0.38895983423268921, 0.43156050120631062, 0.18784825518278428, 0.23958644581530344, 0.16386055449534453, 0.42697777787951602, 0.07748674945294963, 0.47881248869131277, 0.37267534771232319, 0.8083924735050152, 0.29179189019925428, 0.29802029242077777, 0.2054027587360118, 0.10766996738143179, 0.39134998593956405, 0.36412616130029274, 0.37015995608450686, 0.61246049427537563, 0.59897086243095388, 0.20718454055122912, 0.14334918487088333, 0.42189815231899974, 0.21738749430714899, 0.39304902005163428, 0.50261637732761, 0.20759334134444557, 0.21005124432686503, 0.31027042275886835, 0.71335371670327341, 1.4192781030245714, 0.50672517580861098, 0.18067653694488042, 0.85952730574991043, 0.19249388587333111,
                  0.18826477050167958, 0.22742885130411533, 0.24125995809534906, 0.045750800392306591, 0.78242626285998884, 0.20614461737324383, 0.56783904985512668, 0.33500622312674566, 0.043533317315170454, 0.27874382197104552, 0.3685525858770754, 0.1751812314517863, 0.2532293526642409, 0.15785104775566922, 0.2390711833181299, 0.42911409505776471, 0.16819203200742916, 0.40372196988518594, 0.43241512178368696, 0.30424021778439686, 0.19085162018033855, 0.58462246847853661, 0.631050399423982, 0.30137309454374051, 0.27565096287611918, 0.33033553618821287, 0.47665164689105288, 0.34084703029218633, 0.27978844627773986, 0.010536582324145049, 0.050935298127348511, 0.23536721808983668, 0.22364077067346355, 0.31704429093519465, 1.0296141286403104, 0.38123158028252929, 0.27538594123938104, 0.81446088474774558, 0.2660327021825486, 0.195234318277725, 0.462999083371401, 1.0587870384537303, 0.40260375543813692, 0.39471961997665139, 0.29845505700189406, 1.0259474557694457, 0.52111381852233729, 0.29182304834212835, 0.34045457181768657, 0.20417518807825608]

        result = AnnotatorResult(
            "test", "test", "na", observed, values, reference=None, pseudo_count=0)

        self.assertEqual(result.pvalue, 0.57)


class TestSamples(GatTest):

    nsamples = 1000
    ntracks = 50
    nsegments = 10000
    nisochores = 20

    def testDelete(self):
        '''test to track down a memory leak.'''
        # from guppy import hpy
        # hp = hpy()
        # hp.setrelheap()
        return

        samples = Samples()

        for track in range(self.ntracks):
            track_id = str(track)
            # print track_id
            for sample in range(self.nsamples):
                sample_id = str(sample)
                for isochore in range(self.nisochores):
                    isochore_id = str(isochore)
                    r = SegmentList(allocate=self.nsegments)
                    samples.add(track_id, sample_id, isochore_id, r)

            del samples[track_id]
            # print len(samples)
            # h = hp.heap()

            # print h


class TestCaching(GatTest):

    sample_size = 10
    workspace_size = 1000

    def testCaching(self):
        return
        workspaces, segments, annotations = \
            IntervalCollection( "workspace" ), \
            IntervalCollection( "segment" ), \
            IntervalCollection("annotation")

        workspaces.add("default", "chr1", SegmentList(iter=[(0, self.workspace_size), ],
                                                      normalize=True))
        workspace = workspaces["default"]

        segments.add("default", "chr1", SegmentList(iter=[(0, 1), ],
                                                    normalize=True))

        # annotations: a collection of segments with increasing density
        # all are overlapping the segments
        for y in range(1, 100, 2):
            annotations.add("%03i" % y, "chr1",
                            SegmentList(iter=[(0, y), ],
                                        normalize=True))

        workspace_size = workspace["chr1"].sum()

        sampler = SamplerAnnotator(bucket_size=1, nbuckets=self.workspace_size)

        if os.path.exists("test.cache"):
            os.remove("test.cache")

        outsamples = SamplesCached("test.cache")
        saved_samples = {}

        for track in segments.tracks:
            segs = segments[track]
            for x in range(self.sample_size):
                for isochore in list(segs.keys()):
                    r = sampler.sample(segs[isochore], workspace[isochore])
                    saved_samples[(track, x, isochore)] = r
                    outsamples.add(track, x, isochore, r)

        del outsamples

        insamples = SamplesCached("test.cache")

        for track in segments.tracks:
            segs = segments[track]
            for x in range(self.sample_size):
                for isochore in list(segs.keys()):
                    insamples.load(track, x, isochore)

        # compare
        for track in segments.tracks:
            segs = segments[track]
            for x in range(self.sample_size):
                for isochore in list(segs.keys()):
                    self.assertEqual(saved_samples[(track, x, isochore)].asList(),
                                     insamples[track][x][isochore].asList())

        if os.path.exists("test.cache"):
            os.remove("test.cache")
        if os.path.exists("test.cache.idx"):
            os.remove("test.cache.idx")


class TestStats(GatTest):

    ntracks = 10  # 17
    nannotations = 10  # 90
    nsamples = 1000

    def testPValueComputation(self):

        annotation_size = 100
        workspace_size = 1000
        segment_size = 10

        l = 10

        for y in range(1, l):

            samples = [1] * y + [0] * (l - y)

            for x, s in enumerate(samples):

                g = AnnotatorResult("track",
                                    "samples",
                                    'counter',
                                    s,
                                    samples)
                self.assertEqual(
                    g.isSampleSignificantAtPvalue(x, g.pvalue), True)

                t = 0
                for z, s2 in enumerate(samples):
                    t += g.isSampleSignificantAtPvalue(z, g.pvalue)
                fpr = float(t) / l

                # == should work, but there is a problem
                # for pvalue = 0.5
                self.assertTrue(fpr >= g.pvalue - 0.0001,
                             "fpr (%f) != pvalue (%f): y=%i, x=%i, s=%i, t=%i" %
                             (fpr, g.pvalue, y, x, s, t))

    def testPValueComputation2(self):

        samples = [0] * 66 + [1] * 2 + [2] * 20 + [3] * \
            1 + [4] * 6 + [6] * 2 + [8] * 2 + [16] * 1

        obs = 16

        g = AnnotatorResult("track", "samples",
                            'counter',
                            obs,
                            samples)

        self.assertEqual(g.pvalue, 0.01)

    def testStats(self):

        annotation_size = 100
        workspace_size = 1000
        segment_size = 10

        observed = numpy.random.hypergeometric(annotation_size,
                                               workspace_size - annotation_size,
                                               segment_size,
                                               self.ntracks * self.nannotations)

        results = []
        x = 0
        for track in range(self.ntracks):
            for annotation in range(self.nannotations):
                samples = numpy.random.hypergeometric(annotation_size,
                                                      workspace_size - annotation_size,
                                                      segment_size,
                                                      self.nsamples)

                samples.sort()
                # print "obs", observed[x], "sample=", samples

                results.append(AnnotatorResult(str(track),
                                               str(annotation),
                                               'counter',
                                               observed[x],
                                               samples))

                x += 1

        computeFDR(results)
        for r in results:
            self.assertTrue(r.qvalue > 0.5, "%f" % r.qvalue)

if __name__ == '__main__':
    unittest.main()



from posix.types cimport off_t
from SegmentList cimport Position, Segment, SegmentList
from CoordinateList cimport CoordinateList

cdef extern from "gat_utils.h":
    long searchsorted(void * base,
                       size_t nmemb,
                       size_t size,
                       void * target,
                       int(*compar)(void *, void *))

    long searchargsorted(void * base,
                         int * sorted,
                         size_t nmemb,
                         size_t size,
                         void * target,
                         int(*compar)(void *, void *))


cdef class PositionList(CoordinateList):

    cdef Position * positions
    cdef size_t npositions
    cdef size_t allocated
    cdef bint is_sorted	
    cdef bint is_normalized

    cpdef sort(self)
    cpdef normalize(self)
    cpdef add(self, Position pos)
    cpdef Position max(self)
    cpdef Position sum(self)
    cpdef Position min(self)
    cpdef clear(self)
    cpdef Position overlapWithRange(
        self, Position start, Position end)
    cpdef Position intersectionWithSegments(
        self, SegmentList other, mode = *)
    cpdef void intersect(self, SegmentList other)
    cpdef PositionList clone(self)

    # C only methods
    cdef _resize(self, int npositions)
    cdef _add(self, Position pos)
    cdef int _getInsertionPoint(self, Position other)
    cdef Position overlap(self, Segment other)

    # Functions overloaded from CoordinateList
    cdef off_t toMMAP(self, void *, int, off_t)
    cdef void fromMMAP(self)



"""test gat implementation."""

import unittest
import random
import pickle

from gat.SegmentList import SegmentList


class GatTest(unittest.TestCase):

    def shortDescription(self):
        return None


class TestSegmentList(GatTest):
    '''test segment list implementation.'''

    def testCreateAndClear(self):
        s = SegmentList()
        self.assertEqual(0, len(s))
        s.add(0, 100)
        self.assertEqual(1, len(s))
        s.clear()
        self.assertEqual(0, len(s))

    def testNormalize1(self):
        '''non-overlapping segments.'''

        ss = [(x, x + 10) for x in range(0, 1000, 100)]
        random.shuffle(ss)
        s = SegmentList()
        for start, end in ss:
            s.add(start, end)
        s.normalize()

        self.assertEqual(len(s), 10)
        self.assertEqual(s.sum(), 100)
        s2 = SegmentList(iter=ss)
        s2.merge(-1)
        self.assertEqual(s, s2)

    def testNormalize1b(self):
        '''non-overlapping segments.'''

        ss = [(x, x + 10) for x in range(100, 1100, 100)]
        random.shuffle(ss)
        s = SegmentList()
        for start, end in ss:
            s.add(start, end)
        s.normalize()
        self.assertEqual(len(s), 10)
        self.assertEqual(s.sum(), 100)
        s2 = SegmentList(iter=ss)
        s2.merge(-1)
        self.assertEqual(s, s2)

    def testNormalizeEmpty(self):
        '''non-overlapping segments.'''

        s = SegmentList()
        self.assertEqual(len(s), 0)
        s.normalize()
        self.assertEqual(len(s), 0)
        self.assertEqual(s.isNormalized, 1)
        s2 = SegmentList()
        s2.merge(-1)
        self.assertEqual(s, s2)

    def testNormalizeEmptySegment(self):
        s = SegmentList(iter=[(0, 0), ])
        s.normalize()
        self.assertEqual(s.isNormalized, 1)
        self.assertEqual(len(s), 0)

        s = SegmentList(iter=[(0, 0), (0, 0)])
        s.normalize()
        self.assertEqual(s.isNormalized, 1)
        self.assertEqual(len(s), 0)

        ss = [(0, 0), (0, 1), (0, 2), (0, 3), (0, 4),
              (0, 5), (0, 6), (0, 7), (0, 8), (0, 9)]
        s = SegmentList(iter=ss)
        s.normalize()
        self.assertEqual(s.isNormalized, 1)
        self.assertEqual(len(s), 1)

        s2 = SegmentList(iter=ss)
        s2.merge(-1)
        self.assertEqual(s, s2)

    def testNormalize2(self):
        '''overlapping segments.'''

        ss = [(x, x + 1000) for x in range(0, 1000, 100)]
        random.shuffle(ss)
        s = SegmentList()
        for start, end in ss:
            s.add(start, end)
        s.normalize()
        self.assertEqual(len(s), 1)
        self.assertEqual(s.sum(), 1900)

    def testNormalize3(self):
        '''non-overlapping but adjacent segments.'''

        ss = [(x, x + 100) for x in range(0, 1000, 100)]
        random.shuffle(ss)
        s = SegmentList()
        for start, end in ss:
            s.add(start, end)
        s.normalize()

        self.assertEqual(len(s), 10)
        self.assertEqual(s.sum(), 1000)
        s2 = SegmentList(iter=ss)
        s2.merge(-1)
        self.assertEqual(s, s2)

    def testNormalize4(self):
        # test multiple interleaved segments
        ss = [(x, x + 100) for x in range(0, 1000, 10)]
        s = SegmentList()
        for start, end in ss:
            s.add(start, end)
        s.normalize()
        self.assertEqual(len(s), 1)
        self.assertEqual(s.sum(), 1090)
        s2 = SegmentList(iter=ss)
        s2.merge(-1)
        self.assertEqual(s, s2)

    def testNormalize5(self):
        ss = [(489, 589), (1966, 2066), (2786, 2886), (0, 0), (3889, 3972), (3998, 4098),
              (6441, 6541), (6937, 7054), (7392, 7492), (8154, 8254), (9046, 9146)]
        s = SegmentList(iter=ss)
        s.normalize()
        self.assertEqual(len(s), len([x for x in ss if x[1] - x[0] > 0]))
        self.assertEqual(s.sum(), 1000)
        s2 = SegmentList(iter=ss)
        s2.merge(-1)
        self.assertEqual(s, s2)

    def testExtend(self):

        s1 = SegmentList(iter=[(x, x + 100) for x in range(0, 1000, 100)])
        s2 = SegmentList(iter=[(x, x + 100) for x in range(2000, 3000, 100)])
        s1.extend(s2)
        self.assertEqual(s1.sum(), s2.sum() * 2)
        self.assertEqual(len(s1), len(s2) * 2)

    def testTrim(self):
        '''test trimming over full range of insertion points and deletions.'''

        for point in range(0, 1000):
            for size in range(0, 300):
                ss = [(x, x + 100) for x in range(0, 1000, 100)]
                s = SegmentList(iter=ss,
                                normalize=True)
                orig = s.sum()
                s.trim(point, size)
                self.assertEqual(orig - size, s.sum(),
                                 "trimming error at %i:%i: expected %i, got %i, %s" %
                                 (point, size,
                                  orig - size, s.sum(), str(s)))

    def testInsertionPoint(self):
        '''check insertion point for normalized segment lists'''
        ss = [(x, x + 10) for x in range(0, 100, 10)]
        s = SegmentList(iter=ss,
                        normalize=True)

        for point in range(0, 100):
            p = s.getInsertionPoint(point, point + 1)
            self.assertEqual(p, point // 10)

        ss = [(x, x + 10) for x in range(0, 100, 20)]
        s = SegmentList(iter=ss,
                        normalize=True)

        for point in range(0, 100):
            p = s.getInsertionPoint(point, point + 1)
            if point >= 90:
                self.assertEqual(p, len(s))
            else:
                self.assertEqual(p, point // 20)

        ss = [(x, x + 10) for x in range(10, 100, 20)]
        s = SegmentList(iter=ss,
                        normalize=True)

        for point in range(0, 100):
            p = s.getInsertionPoint(point, point + 1)
            self.assertEqual(p, (point - 10) // 20)

    def testInsertionPointNonNormalized(self):
        '''check insertion point for unnormalized segment lists.'''
        ss = [(x, x + 20) for x in range(0, 100, 10)]
        s = SegmentList(iter=ss,
                        normalize=False)

        for point in range(0, 100):
            self.assertRaises(
                AssertionError, s.getInsertionPoint, point, point + 1)

    def testMergeAdjacent(self):
        ss = [(x, x + 100) for x in range(0, 1000, 100)]
        random.shuffle(ss)
        s = SegmentList(iter=ss)
        s.merge(0)
        self.assertEqual(len(s), 1)
        self.assertEqual(s.sum(), 1000)

    def testMergeNeighbours(self):

        for y in range(0, 5):
            ss = [(x, x + 100 - y) for x in range(0, 1000, 100)]
            random.shuffle(ss)
            for x in range(0, y + 1):
                s = SegmentList(iter=ss)
                s.merge(x)
                if x < y:
                    self.assertEqual(len(s), 10)
                    self.assertEqual(s.sum(), 1000 - 10 * y)
                else:
                    self.assertEqual(len(s), 1)
                    self.assertEqual(s.sum(), 1000 - y)

    def testExpand(self):
        ss = [(x, x + 10) for x in range(0, 1000, 100)]
        s = SegmentList(iter=ss, normalize=True)
        s.expand_segments(2.0)
        self.assertEqual(s.sum(), 195)
        self.assertEqual(len(s), 10)

    def testLargest(self):
        ss = [(x, x + x / 10) for x in range(0, 1000, 100)]
        s = SegmentList(iter=ss, normalize=True)
        p = s.largest()
        self.assertEqual(p['end'] - p['start'], 90)

    def testGetFilledSegmentsFromStart(self):
        ss = [(x, x + 10) for x in range(0, 120, 20)]
        s = SegmentList(iter=ss, normalize=True)
        for x in range(0, 120, 5):
            f = s.getFilledSegmentsFromStart(x, 20)
            self.assertEqual(f.sum(), 20)
            if x >= 110:
                self.assertEqual(f.min(), s.min())
                self.assertEqual(f.max(), 30)
            elif x > 80:
                self.assertEqual(f.min(), s.min())
                self.assertEqual(f.max(), s.max())
            else:
                if x in (0, 20, 40, 60, 80):
                    self.assertEqual(f.min(), x)
                    self.assertEqual(f.max(), x + 30)
                elif x in (10, 30, 50, 70):
                    self.assertEqual(f.min(), x + 10)
                    self.assertEqual(f.max(), x + 40)
                elif x in (5, 25, 45, 65):
                    self.assertEqual(f.min(), x)
                    self.assertEqual(f.max(), x + 40)
                elif x in (15, 35, 55, 75):
                    self.assertEqual(f.min(), x + 5)
                    self.assertEqual(f.max(), x + 35)
            self.assertEqual(
                s.sum(), s.getFilledSegmentsFromStart(x, 100).sum())

    def testGetFilledSegmentsFromEnd(self):
        ss = [(x, x + 10) for x in range(0, 120, 20)]
        s = SegmentList(iter=ss, normalize=True)
        for x in range(0, 120, 5):
            f = s.getFilledSegmentsFromEnd(x, 20)
            self.assertEqual(f.sum(), 20)
            if x == 0:
                self.assertEqual(f.max(), s.max())
                self.assertEqual(f.min(), 80)
            elif x < 30:
                self.assertEqual(f.max(), s.max())
                self.assertEqual(f.min(), s.min())
            else:
                if x in (40, 60, 80, 100):
                    self.assertEqual(f.min(), x - 40)
                    self.assertEqual(f.max(), x - 10)
                elif x in (30, 50, 70, 90):
                    self.assertEqual(f.min(), x - 30)
                    self.assertEqual(f.max(), x)
                elif x in (45, 65, 85):
                    self.assertEqual(f.min(), x - 40)
                    self.assertEqual(f.max(), x)
                elif x in (35, 55, 75, 95):
                    self.assertEqual(f.min(), x - 35)
                    self.assertEqual(f.max(), x - 5)

            self.assertEqual(s.sum(), s.getFilledSegmentsFromEnd(x, 100).sum())

    def testPickling(self):
        ss = [(x, x + 10) for x in range(0, 120, 20)]
        s = SegmentList(iter=ss, normalize=True)

        b = pickle.loads(pickle.dumps(s))

        self.assertEqual(s, b)

    def testSharing(self):

        ss = [(x, x + 10) for x in range(0, 120, 20)]
        s = SegmentList(iter=ss, normalize=True)
        s.share("/testshare")
        n = SegmentList(share=s)
        self.assertEqual(s, n)

    def testPickledSharing(self):

        ss = [(x, x + 10) for x in range(0, 120, 20)]
        s = SegmentList(iter=ss, normalize=True)
        s.share("/testshare")
        b = pickle.loads(pickle.dumps(s))
        self.assertEqual(s, b)

    def testUnsharing(self):

        ss = [(x, x + 10) for x in range(0, 120, 20)]
        s = SegmentList(iter=ss, normalize=True)
        s.share("/testshare")
        s.unshare()
        s.share("/testshare")
        s.unshare()


class TestSegmentListOverlap(GatTest):

    def setUp(self):
        self.a = SegmentList(iter=((x, x + 10)
                                   for x in range(0, 1000, 100)), normalize = True)

    def testOverlapFull(self):
        self.assertEqual(self.a.overlapWithRange(0, 1000), self.a.sum())

    def testOverlapHalf(self):
        self.assertEqual(self.a.overlapWithRange(0, 500), self.a.sum() / 2)
        self.assertEqual(self.a.overlapWithRange(500, 1000), self.a.sum() / 2)

    def testOverlapAfter(self):
        self.assertEqual(self.a.overlapWithRange(900, 910), 10)
        self.assertEqual(self.a.overlapWithRange(905, 915), 5)

    def testOverlapNone(self):
        self.assertEqual(self.a.overlapWithRange(1000, 2000), 0)
        self.assertEqual(self.a.overlapWithRange(2000, 3000), 0)

    def testOverlapOutOfRange(self):
        self.assertRaises(OverflowError, self.a.overlapWithRange, -100, 5)
        self.assertRaises(OverflowError, self.a.overlapWithRange, -100, -50)
        self.assertEqual(self.a.overlapWithRange(905, 1100), 5)
        self.assertEqual(self.a.overlapWithRange(1000, 1100), 0)

    def testOverlapAll(self):
        for x in range(0, 1000, 100):
            for y in range(0, 10):
                self.assertEqual(self.a.overlapWithRange(x + y, x + y + 1), 1,
                                 "no overlap failure at %i: %i" % (x + y, self.a.overlapWithRange(x + y, x + y + 1)))
            for y in range(10, 100):
                self.assertEqual(self.a.overlapWithRange(
                    x + y, x + y + 1), 0, "overlap failure at %i" % (x + y))


class TestSegmentListIntersection(GatTest):

    def setUp(self):
        self.a = SegmentList(
            iter=((x, x + 10) for x in range(0, 1000, 100)), normalize=True)

    def testIntersectionFull(self):
        b = SegmentList(iter=[(0, 1000)], normalize=True)
        b.intersect(self.a)
        self.assertEqual(b.asList(), self.a.asList())

    def testIntersectionSelf(self):
        self.a.intersect(self.a)
        self.assertEqual(self.a.asList(), self.a.asList())

    def testIntersectionCopy(self):
        b = SegmentList(clone=self.a)
        b.intersect(self.a)
        self.assertEqual(b.asList(), self.a.asList())

    def testNoIntersection(self):
        b = SegmentList(iter=((x, x + 10)
                              for x in range(10, 1000, 100)), normalize = True)
        b.intersect(self.a)
        self.assertEqual(b.asList(), [])
        self.assertEqual(b.isEmpty, True)

    def testPartialIntersection(self):
        b = SegmentList(iter=((x, x + 10)
                              for x in range(5, 1000, 100)), normalize = True)
        b.intersect(self.a)
        self.assertEqual(len(b), len(self.a))
        self.assertEqual(b.sum(), self.a.sum() / 2)

    def testOverlap(self):
        '''test if number of segments intersection is correct.'''

        b = SegmentList(iter=((x, x + 10)
                              for x in range(5, 1000, 100)), normalize = True)
        self.assertEqual(self.a.intersectionWithSegments(b), len(b))
        self.assertEqual(b.intersectionWithSegments(self.a), len(b))

        # no intersection
        b = SegmentList(iter=((x, x + 10)
                              for x in range(10, 1000, 100)), normalize = True)
        self.assertEqual(self.a.intersectionWithSegments(b), 0)
        self.assertEqual(b.intersectionWithSegments(self.a), 0)

        # double the number of segments in b
        b = SegmentList(iter=[(x, x + 5) for x in range(0, 1000, 100)] +
                        [(x + 5, x + 10) for x in range(0, 1000, 100)],
                        normalize=True)
        self.assertEqual(self.a.intersectionWithSegments(b), 10)
        self.assertEqual(b.intersectionWithSegments(self.a), 20)

    def testFilter(self):

        b = SegmentList(iter=((x, x + 5)
                              for x in range(500, 2000, 100)), normalize=True)
        b.filter(self.a)
        self.assertEqual(b.asList(), [
            (500, 505), (600, 605), (700, 705), (800, 805),
            (900, 905)])

        b = SegmentList(iter=((0, 56), ))
        c = SegmentList(iter=[(0, 50), (75, 125)])
        b.filter(c)
        self.assertEqual(b.asList(), [(0, 56)])

        b = SegmentList(iter=((0, 56), ))
        c = SegmentList(iter=[(0, 10)])
        b.filter(c)
        self.assertEqual(b.asList(), [(0, 56)])


class TestSegmentListSubtract(GatTest):

    def setUp(self):
        self.a = SegmentList(
            iter=((x, x + 10) for x in range(0, 1000, 100)), normalize=True)

    def testCompleteOverlap(self):
        b = SegmentList(iter=[(0, 1000)], normalize=True)
        b.subtract(self.a)
        c = [(10, 100), (110, 200), (210, 300), (310, 400), (410, 500),
             (510, 600), (610, 700), (710, 800), (810, 900), (910, 1000)]
        self.assertEqual(b.asList(), c)

    def testFullSubtraction(self):
        b = SegmentList(iter=[(0, 1000)], normalize=True)
        self.a.subtract(b)
        self.assertEqual(len(self.a), 0)

    def testSelfSubtraction(self):
        self.a.subtract(self.a)
        self.assertEqual(len(self.a), 0)

    def testSameSubtraction(self):
        b = SegmentList(clone=self.a)
        b.subtract(self.a)
        self.assertEqual(len(b), 0)

    def testOverlap(self):

        b = SegmentList(iter=((x, x + 10)
                              for x in range(5, 1000, 100)), normalize = True)
        b.subtract(self.a)
        c = [(10, 15), (110, 115), (210, 215), (310, 315), (410, 415),
             (510, 515), (610, 615), (710, 715), (810, 815), (910, 915)]
        self.assertEqual(b.asList(), c)

    def testSingleSegmentSubtraction(self):
        a = SegmentList(iter=[(0, 12000)], normalize=True)
        b = SegmentList(iter=[(0, 10000)], normalize=True)
        a.subtract(b)
        self.assertEqual(a.asList(), [(10000, 12000)])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
##########################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
##########################################################################
'''
gat-compare - compare two gat runs
==================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script compares gat runs and compares statistical significance of
fold change differences between two or more regions of interest.

This script requires counts files created by gat (option
``--output-counts-pattern``).

The reported observed value between sets 1 and 2 is the difference of
fc2 and fc1:

   * observed = fold2 - fold1

The script accepts one or more output files from a gat run.

If only a single file is given, the significance of fold change
differences are computed for each pairwise comparison of
annotations. Thus, the script might answer the question if a
transcription factor is more enriched in promotors than in enhancers.

If multiple files are given, the script computes the fold change
differences for the same annotation for all pairwise combinations of
:term:`segments of interest` across the different files. Thus, the
script might answer the question if transcription factor A is more
enriched in promotors than transcription factor B is enriched in
promotors.

Usage
-----

Example::

   python gat-compare.py tfA.counts.tsv.gz tfB.counts.tsv.gz
      
Type::

   python gat-compare.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os
import sys
import re
import optparse
import collections
import types
import glob
import time
import math
import itertools
import numpy

import gat
import gat.Experiment as E
import gat.IOTools as IOTools
import gat.IO as IO
import gat.Engine as Engine


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = optparse.OptionParser(version="%prog version: $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $",
                                   usage=globals()["__doc__"])

    parser.add_option("-o", "--order", dest="output_order", type="choice",
                      choices=(
                          "track", "annotation", "fold", "pvalue", "qvalue", "observed"),
                      help="order results in output by fold, track, etc. [default=%default].")

    parser.add_option("-p", "--pvalue-method", dest="pvalue_method", type="choice",
                      choices=("empirical", "norm", ),
                      help="type of pvalue reported [default=%default].")

    parser.add_option("-q", "--qvalue-method", dest="qvalue_method", type="choice",
                      choices=(
                          "storey", "BH", "bonferroni", "holm", "hommel", "hochberg", "BY", "none"),
                      help="method to perform multiple testing correction by controlling the fdr [default=%default].")

    parser.add_option("--qvalue-lambda", dest="qvalue_lambda", type="float",
                      help="fdr computation: lambda [default=%default].")

    parser.add_option("--qvalue-pi0-method", dest="qvalue_pi0_method", type="choice",
                      choices=("smoother", "bootstrap"),
                      help="fdr computation: method for estimating pi0 [default=%default].")

    parser.add_option("--descriptions", dest="input_filename_descriptions", type="string",
                      help="filename mapping annotation terms to descriptions. "
                      " if given, the output table will contain additional columns "
                      " [default=%default]")

    parser.add_option("--pseudo-count", dest="pseudo_count", type="float",
                      help="pseudo count. The pseudo count is added to both the observed and expected overlap. "
                      " Using a pseudo-count avoids gat reporting fold changes of 0 [default=%default].")

    parser.add_option("--output-plots-pattern", dest="output_plots_pattern", type="string",
                      help="output pattern for plots [default=%default]")

    parser.set_defaults(
        pvalue_method="empirical",
        qvalue_method="BH",
        qvalue_lambda=None,
        qvalue_pi0_method="smoother",
        # pseudo count for fold change computation to avoid 0 fc
        pseudo_count=1.0,
        output_order="observed",
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv, add_output_options=True)

    input_filenames_counts = args

    ##################################################
    E.info("received %i filenames with counts" % len(input_filenames_counts))

    ##################################################
    description_header, descriptions, description_width = IO.readDescriptions(
        options)

    all_annotator_results = []

    for input_filename_counts in input_filenames_counts:

        E.info("processing %s" % input_filename_counts)

        annotator_results = gat.fromCounts(input_filename_counts)

        ##################################################
        if options.pvalue_method != "empirical":
            E.info("updating pvalues to %s" % options.pvalue_method)
            Engine.updatePValues(annotator_results, options.pvalue_method)

        ##################################################
        ##################################################
        ##################################################
        # compute global fdr
        ##################################################
        E.info("computing FDR statistics")
        Engine.updateQValues(annotator_results,
                                method=options.qvalue_method,
                                vlambda=options.qvalue_lambda,
                                pi0_method=options.qvalue_pi0_method)

        all_annotator_results.append(annotator_results)

    pseudo_count = options.pseudo_count
    results = []

    if len(all_annotator_results) == 1:
        E.info("performing pairwise comparison within a single file")

        # collect all annotations
        annotations, segments = list(), set()
        for x in all_annotator_results[0]:
            segments.add(x.track)
            annotations.append(x)

        if len(segments) != 1:
            raise NotImplementedError("multiple segments of interest")

        for data1, data2 in itertools.combinations(annotations, 2):

            # note that fold changes can be very large if there are 0 samples
            # this is fine for getting the distributional params (mean, stddev)
            fold_changes1 = data1.observed / (data1.samples + pseudo_count)
            fold_changes2 = data2.observed / (data2.samples + pseudo_count)

            # add a separate fc pseudo-count to avoid 0 values
            fold_changes1 += 0.0001
            fold_changes2 += 0.0001

            # Test is if relative fold change rfc is different from 1
            # note: rfc = fc1 / fc2 = obs1 / exp1 * obs2 / exp2
            #                       = obs1 / obs2 * exp2 / exp1
            # Thus, it is equivalent to test rfc = obs1/obs2 versus exp2 / exp1
            #
            # Convert to log space for easier plotting
            # Move the observed fold ratio in order to get an idea of the magnitude
            # of the underlying fold change
            delta_fold = data2.fold - data1.fold
            sampled_delta_fold = numpy.log(
                fold_changes1 / fold_changes2) + delta_fold
            observed_delta_fold = 0.0 + delta_fold

            result = Engine.AnnotatorResult(data1.annotation, data2.annotation,
                                               "na",
                                               observed_delta_fold,
                                               sampled_delta_fold,
                                               reference=None,
                                               pseudo_count=0)

            results.append(result)

    else:
        E.info("performing pairwise comparison between multiple files")

        ##################################################
        # perform pairwise comparison
        for index1, index2 in itertools.combinations(range(len(input_filenames_counts)), 2):
            E.info("comparing %i and %i" % (index1, index2))
            a, b = all_annotator_results[index1], all_annotator_results[index2]

            # index results in a and b
            aa = collections.defaultdict(dict)
            for x in a:
                aa[x.track][x.annotation] = x

            bb = collections.defaultdict(dict)
            for x in b:
                bb[x.track][x.annotation] = x
            
            tracks_a = set(aa.keys())
            tracks_b = set(bb.keys())
            shared_tracks = tracks_a.intersection(tracks_b)
            if len(shared_tracks) == 0:
                E.warn("no shared tracks between {} and {}".format(
                        index1, index2))
                
            for track in sorted(shared_tracks):
                E.debug("computing results for track {}".format(track))
                # get shared annotations
                annotations1 = aa[track].keys()
                annotations2 = bb[track].keys()
                shared_annotations = list(
                    set(annotations1).intersection(set(annotations2)))
                E.info("%i shared annotations" % len(shared_annotations))

                for annotation in shared_annotations:

                    # if not annotation.startswith("Ram:"): continue

                    data1 = aa[track][annotation]
                    data2 = bb[track][annotation]

                    # note that fold changes can be very large if there are 0 samples
                    # this is fine for getting the distributional params (mean,
                    # stddev)
                    fold_changes1 = data1.observed / (data1.samples + pseudo_count)
                    fold_changes2 = data2.observed / (data2.samples + pseudo_count)

                    # add a separate fc pseudo-count to avoid 0 values
                    fold_changes1 += 0.0001
                    fold_changes2 += 0.0001

                    # Test is if relative fold change rfc is different from 1
                    # note: rfc = fc1 / fc2 = obs1 / exp1 * obs2 / exp2
                    #                       = obs1 / obs2 * exp2 / exp1
                    # Thus, it is equivalent to test rfc = obs1/obs2 versus exp2 / exp1
                    #
                    # Convert to log space for easier plotting
                    # Move the observed fold ratio in order to get an idea of the magnitude
                    # of the underlying fold change
                    delta_fold = data2.fold - data1.fold
                    sampled_delta_fold = numpy.log(
                        fold_changes1 / fold_changes2) + delta_fold
                    observed_delta_fold = 0.0 + delta_fold

                    result = Engine.AnnotatorResult(track,
                                                       annotation,
                                                       "na",
                                                       observed_delta_fold,
                                                       sampled_delta_fold,
                                                       reference=None,
                                                       pseudo_count=0)

                    results.append(result)

    if len(results) == 0:
        E.critical("no results found")
        E.Stop()
        return

    IO.outputResults(results,
                     options,
                     Engine.AnnotatorResult.headers,
                     description_header,
                     description_width,
                     descriptions,
                     format_observed="%6.4f")

    IO.plotResults(results, options)

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



import re
import os
import glob
import gat
import gat.IOTools as IOTools
import gat.Experiment as E
import gat.Stats as Stats
import numpy

import gat.SegmentList as SegmentList
import gat.Engine as Engine

try:
    import matplotlib.pyplot as plt
    HASPLOT = True
except (ImportError, RuntimeError):
    HASPLOT = False


def dumpStats(coll, section, options):
    if section in options.output_stats or \
            "all" in options.output_stats or \
            len([x for x in options.output_stats
                 if re.search(x, section)]) > 0:
        coll.outputStats(E.openOutputFile(section))


def dumpBed(coll, section, options):
    if section in options.output_bed or \
            "all" in options.output_bed or \
            len([x for x in options.output_bed if re.search(x, section)]) > 0:
        coll.save(E.openOutputFile(section + ".bed"))


def readSegmentList(label,
                    filenames,
                    enable_split_tracks=False,
                    ignore_tracks=False):
    """read one or more segment files.

    Arguments
    ---------
    label : string
        Label to use for IntervalCollection.
    filenames : list
        List of filenames to load in :term:`bed` format.
    enable_split_tracks : bool
        If True, allow tracks to be split across multiple files.
    ignore_tracks : int
        If True, ignore track information.

    Returns
    -------
    segments : IntervalCollection
        The segment collection.
    """
    results = Engine.IntervalCollection(name=label)
    E.info("%s: reading tracks from %i files" % (label, len(filenames)))
    results.load(filenames,
                 allow_multiple=enable_split_tracks,
                 ignore_tracks=ignore_tracks)
    E.info("%s: read %i tracks from %i files" %
           (label, len(results), len(filenames)))
    return results


def readAnnotatorResults(filename):
    '''load annotator results from a tab-separated results table.'''

    annotator_results = []

    with IOTools.openFile(filename, "r") as infile:
        for line in infile:
            if line.startswith("#"):
                continue
            if line.startswith("track"):
                continue
            r = gat.DummyAnnotatorResult._fromLine(line)
            annotator_results.append(r)

    return annotator_results


def expandGlobs(infiles):
    return IOTools.flatten([glob.glob(x) for x in infiles])


def buildSegments(options):
    '''load segments, annotations and workspace from parameters
    defined in *options*.

    The workspace will be split by isochores.

    returns segments, annotations and workspace.
    '''

    options.segment_files = expandGlobs(options.segment_files)
    options.annotation_files = expandGlobs(options.annotation_files)
    options.workspace_files = expandGlobs(options.workspace_files)
    options.sample_files = expandGlobs(options.sample_files)

    ##################################################
    # arguments sanity check
    if not options.segment_files:
        raise ValueError("please specify at least one segment file")
    if not options.annotation_files:
        raise ValueError("please specify at least one annotation file")
    if not options.workspace_files:
        raise ValueError("please specify at least one workspace file")

    # read one or more segment files
    segments = readSegmentList("segments",
                               options.segment_files,
                               ignore_tracks=options.ignore_segment_tracks)
    segments.normalize()

    if segments.sum() == 0:
        E.critical("no segments in input file - run aborted")
        raise ValueError("segments file is empty - run aborted")

    if len(segments) > 1000:
        raise ValueError(
            "too many (%i) segment files - use track definitions "
            "or --ignore-segment-tracks" % len(segments))

    annotations = readSegmentList(
        "annotations", options.annotation_files,
        enable_split_tracks=options.enable_split_tracks,
        ignore_tracks=options.annotations_label is not None)

    if options.annotations_label is not None:
        annotations.setName(options.annotations_label)

    if options.annotations_to_points:
        annotations.toPositions(options.annotations_to_points)

    if options.overlapping_annotations:
        # only sort, do not merge
        annotations.sort()
    else:
        annotations.normalize()

    workspaces = readSegmentList(
        "workspaces", options.workspace_files, options,
        options.enable_split_tracks)
    workspaces.normalize()

    # intersect workspaces to build a single workspace
    E.info("collapsing workspaces")
    dumpStats(workspaces, "stats_workspaces_input", options)
    workspaces.collapse()
    dumpStats(workspaces, "stats_workspaces_collapsed", options)

    # use merged workspace only, discard others
    workspaces.restrict("collapsed")

    # build isochores or intersect annotations/segments with workspace
    if options.isochore_files:

        # read one or more isochore files
        isochores = Engine.IntervalCollection(name="isochores")
        E.info("%s: reading isochores from %i files" %
               ("isochores", len(options.isochore_files)))
        isochores.load(options.isochore_files)
        dumpStats(isochores, "stats_isochores_raw", options)

        # merge isochores and check if consistent (fully normalized)
        isochores.sort()

        # check that there are no overlapping segments within isochores
        isochores.check()

        # TODO: flag is_normalized not properly set
        isochores.normalize()

        # check that there are no overlapping segments between isochores

        # truncate isochores to workspace
        # crucial if isochores are larger than workspace.
        isochores.intersect(workspaces["collapsed"])

    else:
        isochores = None

    return segments, annotations, workspaces, isochores


def applyIsochores(segments, annotations, workspaces,
                   options,
                   isochores=None,
                   truncate_segments_to_workspace=False,
                   truncate_workspace_to_annotations=False,
                   restrict_workspace=False,
                   ):
    '''apply isochores to segments and annotations.

    Segments and annotations are filtered in place to keep only those
    overlapping the workspace.

    If *isochores* are given, isochores are applied.

    If *truncate_segments_to_workspace*, truncate segments
    to workspace.

    If *restrict_workspace* is set, the workspace is confined
    to those parts that overlap both a segment and an annotation.

    If *truncate_workspace_to_annotations* is set, the workspace
    is truncated to keep only those parts that overlap annotations.

    returns a workspace divided into isochores.

    '''

    if isochores:
        # intersect isochores and workspaces, segments and annotations
        # workspace and annotations are truncated
        # with segments it is optional.
        E.info("adding isochores to workspace")
        workspaces.toIsochores(isochores, truncate=True)
        annotations.toIsochores(isochores, truncate=True)
        segments.toIsochores(
            isochores, truncate=options.truncate_segments_to_workspace)

        if workspaces.sum() == 0:
            raise ValueError("isochores and workspaces do not overlap")
        if annotations.sum() == 0:
            raise ValueError("isochores and annotations do not overlap")
        if segments.sum() == 0:
            raise ValueError("isochores and segments do not overlap")

        dumpStats(workspaces, "stats_workspaces_isochores", options)
        dumpStats(annotations, "stats_annotations_isochores", options)
        dumpStats(segments, "stats_segments_isochores", options)

        dumpBed(workspaces, "workspaces_isochores", options)
        dumpBed(annotations, "annotations_isochores", options)
        dumpBed(segments, "segments_isochores", options)

    else:
        # intersect workspace and segments/annotations
        # annotations and segments are truncated by workspace
        if options.truncate_segments_to_workspace:
            segments.intersect(workspaces["collapsed"])
        else:
            segments.filter(workspaces["collapsed"])

        annotations.intersect(workspaces["collapsed"])

        dumpStats(annotations, "stats_annotations_truncated", options)
        dumpStats(segments, "stats_segments_truncated", options)

    workspace = workspaces["collapsed"]

    if restrict_workspace:

        E.info("restricting workspace")
        # this is very cumbersome - refactor merge and collapse
        # to return an IntervalDictionary instead of adding it
        # to the list of tracks
        for x in (segments, annotations):
            if "merged" in segments:
                workspace.filter(segments["merged"])
            else:
                segments.merge()
                workspace.filter(segments["merged"])
                del segments["merged"]

        dumpStats(workspaces, "stats_workspaces_restricted", options)

    if truncate_workspace_to_annotations:

        E.info("truncating workspace to annotations")
        annotations.merge()
        annotations["merged"].normalize()
        workspace.intersect(annotations["merged"])
        del annotations["merged"]

        dumpStats(workspaces, "stats_workspaces_truncated", options)

    # segments.dump( open("segments_dump.bed", "w" ) )
    # workspaces.dump( open("workspaces_dump.bed", "w" ) )

    # output overlap stats
    # output segment densities per workspace
    if "overlap" in options.output_stats or \
            "all" in options.output_stats:
        for track in segments.tracks:
            workspaces.outputOverlapStats(
                E.openOutputFile("overlap_%s" % track),
                segments[track])

    return workspace


def readDescriptions(options):
    '''read descriptions from tab separated file.'''

    description_header, descriptions, description_width = [], {}, 0
    if options.input_filename_descriptions:
        E.info("reading descriptions from %s" %
               options.input_filename_descriptions)

        with IOTools.openFile(options.input_filename_descriptions) as inf:
            first = True
            for line in inf:
                if line.startswith("#"):
                    continue
                data = line[:-1].split("\t")

                if description_width:
                    assert len(data) - 1 == description_width, \
                        "inconsistent number of descriptions in %s" %\
                        options.input_filename_descriptions
                else:
                    description_width = len(data) - 1

                if first:
                    description_header = data[1:]
                    first = False
                else:
                    descriptions[data[0]] = data[1:]
        assert len(description_header) == description_width, \
            "number of descriptions (%i) inconsistent with header (%s) in %s" % \
            (description_width, len(description_header),
             options.input_filename_descriptions)

    return description_header, descriptions, description_width


class SegmentsSummary:

    '''summarize segments in a workspace.
    '''
    header = ("all_segments",
              "all_nucleotides",
              "segments_overlapping_workspace",
              "nucleotides_overlapping_workspace",
              "segments_outside_workspace",
              "nucleotides_outside_workspace",
              "truncated_segments",
              "truncated_nucleotides",
              "density_workspace",
              "proportion_truncated_segments",
              "proportion_extending_nucleotides",
              "summary_all_segments",
              "summary_segments_overlapping_workspace",
              "summary_truncated_segments")

    def __init__(self):
        pass

    def update(self, segments, workspace):
        '''compute summary statistics between two segments lists *segments* and *workspace*.
        This method computes:
            * number of segments/nucleotides
            * number of segments/nucleotides overlapping workspace
            * number of segments split by a boundary
            * number of nucleotides extending outside boundary
            * segment size distribution untruncated at boundaries (min, max, mean, media, q1, q3)
            * segment size distributino truncated at boundaries
        '''

        self.all_segments = len(segments)
        self.all_nucleotides = segments.sum()

        # build segments overlapping workspace
        segments_overlapping_workspace = SegmentList.SegmentList(
            clone=segments)
        segments_overlapping_workspace.filter(workspace)

        # build segments truncated by workspace
        truncated_segments = SegmentList.SegmentList(
            clone=segments_overlapping_workspace)
        truncated_segments.intersect(workspace)

        segments_extending_workspace = SegmentList.SegmentList(
            clone=segments)
        segments_extending_workspace.subtract(truncated_segments)

        # compute numbers
        self.segments_overlapping_workspace = len(truncated_segments)
        self.nucleotides_overlapping_workspace = truncated_segments.sum()

        self.segments_outside_workspace = self.all_segments - \
            self.segments_overlapping_workspace
        self.nucleotides_outside_workspace = self.all_nucleotides - \
            self.nucleotides_overlapping_workspace

        self.truncated_segments = len(segments_extending_workspace)
        self.truncated_nucleotides = segments_extending_workspace.sum()

        self.summary_all_segments = segments.summarize()
        self.summary_segments_overlapping_workspace = segments_overlapping_workspace.summarize()
        self.summary_truncated_segments = truncated_segments.summarize()

        workspace_size = workspace.sum()

        self.density_workspace, self.proportion_truncated_segments, self.proportion_extending_nucleotides = 0, 0, 0
        # some quality metrics
        if workspace_size > 0:
            self.density_workspace = float(
                self.nucleotides_overlapping_workspace) / workspace_size
        if self.segments_overlapping_workspace > 0:
            self.proportion_truncated_segments = float(
                self.truncated_segments) / self.segments_overlapping_workspace
            self.proportion_extending_nucleotides = float(
                self.truncated_nucleotides) / segments_overlapping_workspace.sum()

    def __str__(self):
        return "\t".join(["%i" % self.all_segments,
                          "%i" % self.all_nucleotides,
                          "%i" % self.segments_overlapping_workspace,
                          "%i" % self.nucleotides_overlapping_workspace,
                          "%i" % self.segments_outside_workspace,
                          "%i" % self.nucleotides_outside_workspace,
                          "%i" % self.truncated_segments,
                          "%i" % self.truncated_nucleotides,
                          "%f" % self.density_workspace,
                          "%f" % self.proportion_truncated_segments,
                          "%f" % self.proportion_extending_nucleotides,
                          "%s" % str(self.summary_all_segments),
                          "%s" % str(
                              self.summary_segments_overlapping_workspace),
                          "%s" % str(self.summary_truncated_segments)])


def outputMetrics(outfile, segments, workspace, track, section):
    '''output summary metrics

    Outputs summary metrics for segments in workspace. 
    .'''

    stats_per_isochore = []
    for isochore, ss in segments.items():
        stats = SegmentsSummary()
        stats.update(ss, workspace[isochore])
        stats_per_isochore.append(stats)

    for attribute in ("all_segments", "all_nucleotides",
                      "segments_overlapping_workspace",
                      "nucleotides_overlapping_workspace",
                      "nucleotides_outside_workspace",
                      "truncated_segments",
                      "truncated_nucleotides",
                      "density_workspace",
                      "proportion_truncated_segments",
                      "proportion_extending_nucleotides",
                      ):
        values = [getattr(x, attribute) for x in stats_per_isochore]
        outfile.write("%s\t%s\t%s\t%s\n" %
                      (track, section, attribute, Stats.Summary(values)))

    outfile.flush()


def outputResults(results,
                  options,
                  header,
                  description_header,
                  description_width,
                  descriptions,
                  format_observed="%i"):
    '''compute FDR and output results.'''

    pvalues = [x.pvalue for x in results]

    ##################################################
    ##################################################
    ##################################################
    # compute global fdr
    ##################################################
    E.info("computing FDR statistics")
    qvalues = Engine.getQValues(pvalues,
                                   method=options.qvalue_method,
                                   vlambda=options.qvalue_lambda,
                                   pi0_method=options.qvalue_pi0_method)

    try:
        results = [x._replace(qvalue=qvalue)
                   for x, qvalue in zip(results, qvalues)]
        is_tuple = True
    except AttributeError:
        # not a namedtuple
        for x, qvalue in zip(results, qvalues):
            x.qvalue = qvalue
            x.format_observed = format_observed

        is_tuple = False

    counters = set([x.counter for x in results])

    for counter in counters:

        if len(counters) == 1:
            outfile = options.stdout
            output = results
        else:
            outfilename = re.sub("%s", counter, options.output_tables_pattern)
            E.info("output for counter %s goes to outfile %s" %
                   (counter, outfilename))
            outfile = IOTools.openFile(outfilename, "w")
            output = [x for x in results if x.counter == counter]

        outfile.write(
            "\t".join(list(header) + list(description_header)) + "\n")

        if options.output_order == "track":
            output.sort(key=lambda x: (x.track, x.annotation))
        elif options.output_order == "observed":
            output.sort(key=lambda x: x.observed)
        elif options.output_order == "annotation":
            output.sort(key=lambda x: (x.annotation, x.track))
        elif options.output_order == "fold":
            output.sort(key=lambda x: x.fold)
        elif options.output_order == "pvalue":
            output.sort(key=lambda x: x.pvalue)
        elif options.output_order == "qvalue":
            output.sort(key=lambda x: x.qvalue)
        else:
            raise ValueError("unknown sort order %s" % options.output_order)

        for result in output:
            if is_tuple:
                outfile.write("\t".join(map(str, result)))
            else:
                outfile.write(str(result))

            if descriptions:
                try:
                    outfile.write(
                        "\t" + "\t".join(descriptions[result.annotation]))
                except KeyError:
                    outfile.write("\t" + "\t".join([""] * description_width))
            outfile.write("\n")

        if outfile != options.stdout:
            outfile.close()


def plotResults(results, options):
    '''plot annotator results.'''

    ##################################################
    # plot histograms
    if options.output_plots_pattern and HASPLOT:

        def buildPlotFilename(options, key):
            filename = re.sub("%s", key, options.output_plots_pattern)
            filename = re.sub("[^a-zA-Z0-9-_./]", "_", filename)
            dirname = os.path.dirname(filename)
            if dirname and not os.path.exists(dirname):
                os.makedirs(dirname)
            return filename

        E.info("plotting sample stats")

        for r in results:

            plt.figure()
            k = []
            if r.track != "merged":
                k.append(r.track)
            k.append(r.annotation)
            if r.counter != "na":
                k.append(r.counter)
            key = "-".join(k)

            s = r.samples
            hist, bins = numpy.histogram(s,
                                         bins=100)

            # plot bars
            plt.hist(s, bins=100, normed=True, label=key)

            plt.axvline(r.observed, color='r', linewidth=2)

            # plot estimated
            sigma = r.stddev
            mu = r.expected
            plt.plot(bins,
                     1.0 / (sigma * numpy.sqrt(2 * numpy.pi)) *
                     numpy.exp(- (bins - mu) ** 2 / (2 * sigma ** 2)),
                     label="std distribution",
                     linewidth=2,
                     color='g')

            plt.legend()
            filename = buildPlotFilename(options, key)
            plt.savefig(filename)

        E.info("plotting P-value distribution")

        key = "pvalue"
        plt.figure()

        x, bins, y = plt.hist([r.pvalue for r in results],
                              bins=numpy.arange(0, 1.05, 0.025),
                              label="pvalue")

        plt.hist([r.qvalue for r in results],
                 bins=numpy.arange(0, 1.05, 0.025),
                 label="qvalue",
                 alpha=0.5)

        plt.legend()

        # hist, bins = numpy.histogram( \
        #     [r.pvalue for r in Engine.iterator_results(annotator_results) ],
        #     bins = 20 )
        # plt.plot( bins[:-1], hist, label = key )

        filename = buildPlotFilename(options, key)
        plt.savefig(filename)



'''test gat functionality
'''

import unittest
import gat


class TestSNPs(unittest.TestCase):

    def check(self):
        pass

    def testIntervalsPartialOverlap(self):
        '''test with intervals with 
        increasing amount of overlap.

        '''
        return
        workspaces, segments, annotations = \
            gat.IntervalCollection("workspace"), \
            gat.IntervalCollection("segment"), \
            gat.IntervalCollection("annotation")

        workspace_size = 1000

        size = 100

        # workspace of size 1000000
        workspaces.add("default", "chr1",
                       gat.SegmentList(iter=[(0, workspace_size), ],
                                       normalize=True))
        workspace = workspaces["default"]

        # segment of size 10
        segments.add("default", "chr1",
                     gat.SegmentList(iter=[(0, size), ],
                                     normalize=True))

        # annotations: a collection of segments.
        # overlap increases
        annotations.add("full", "chr1",
                        gat.SegmentList(iter=[(y, size + y), ],
                                        normalize=True))

        self.check(workspace, annotations, segments)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
"""\

GAT
***

Genomic annotation tool

"""\

import sys
import glob
import platform

########################################################################
#######################################################################
# Check for dependencies
##
# Is there a way to do this more elegantly?
# 1. Run "pip install numpy"
# 2. Wrap inside functions (works for numpy/pysam, but not cython)
try:
    import numpy
except ImportError:
    raise ImportError(
        "GAT requires numpy to be installed before running setup.py "
        "(pip install numpy)")

try:
    import Cython
except ImportError:
    raise ImportError(
        "GAT code requires cython to be installed before running setup.py "
        "(pip install cython)")

########################################################################
########################################################################
# Import setuptools
# Use existing setuptools, otherwise try ez_setup.
try:
    import setuptools
except ImportError:
    # try to get via ez_setup
    # ez_setup did not work on all machines tested as
    # it uses curl with https protocol, which is not
    # enabled in ScientificLinux
    import ez_setup
    ez_setup.use_setuptools()

from setuptools import setup, Extension

from distutils.version import LooseVersion
if LooseVersion(setuptools.__version__) < LooseVersion('1.1'):
    raise ImportError(
        "GAT requires setuptools 1.1 higher")

###############################################################
# Define dependencies
#
major, minor1, minor2, s, tmp = sys.version_info

if major == 2 and minor1 < 5 or major < 2:
    raise SystemExit("""GAT requires Python 2.5 or later.""")

try:
    from Cython.Distutils import build_ext
except ImportError:
    # no Cython available - use existing C code
    cmdclass = {}
    CoordinateList_sources = ['gat/CoordinateList.c',
                              'utils/gat_utils.c']
    SegmentList_sources = ['gat/SegmentList.c',
                           'utils/gat_utils.c']
    PositionList_sources = ['gat/PositionList.c',
                            'utils/gat_utils.c']
    Engine_sources = ['gat/Engine.c',
                      'utils/gat_utils.c']
else:
    cmdclass = {'build_ext': build_ext}
    CoordinateList_sources = ['gat/CoordinateList.pyx',
                              'utils/gat_utils.c']
    SegmentList_sources = ['gat/SegmentList.pyx',
                           'utils/gat_utils.c']
    PositionList_sources = ['gat/PositionList.pyx',
                            'utils/gat_utils.c']
    Engine_sources = ['gat/Engine.pyx',
                      'utils/gat_utils.c']


install_requires = []

for requirement in (l.strip() for l in open('requirements.txt')
                    if not l.startswith("#")):
    install_requires.append(requirement)


name = "gat"
version = "1.3.6"

classifiers = """
Development Status :: 4 - Beta
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft :: Windows :: Windows NT/2000
Operating System :: OS Independent
Operating System :: POSIX
Operating System :: POSIX :: Linux
Operating System :: Unix
Programming Language :: Python
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Bioinformatics
"""

libraries = ['z']

# link against rt for shared memory access
# macosx has no librt
if platform.system() in ("Linux", "Windows"):
    libraries.append('rt')

CoordinateList = Extension(
    "gat.CoordinateList",
    CoordinateList_sources,
    libraries=libraries,
    library_dirs=[],
    include_dirs=['./utils', numpy.get_include()],
    language="c",
)

SegmentList = Extension(
    "gat.SegmentList",
    SegmentList_sources,
    libraries=libraries,
    library_dirs=[],
    include_dirs=['./utils', numpy.get_include()],
    language="c",
)

PositionList = Extension(
    "gat.PositionList",
    PositionList_sources,
    libraries=libraries,
    library_dirs=[],
    include_dirs=['./utils', numpy.get_include()],
    language="c",
)

Engine = Extension(
    "gat.Engine",
    Engine_sources,
    libraries=libraries,
    library_dirs=[],
    include_dirs=["./utils",
                  "../GatSegmentList",
                  numpy.get_include()],
    language="c",
)

setup(
    name=name,
    version=version,
    description="GenomicAssocationTester",
    long_description=__doc__,
    author="Andreas Heger",
    author_email="andreas.heger@gmail.com",
    license="MIT",
    platforms="ALL",
    url="https://github.com/AndreasHeger/gat",
    py_modules=[x[:-3] for x in glob.glob('gat/*.py')],
    install_requires=install_requires,
    ext_modules=[CoordinateList, SegmentList, PositionList, Engine],
    cmdclass={'build_ext': build_ext},
    scripts=['scripts/gat-run.py',
             'scripts/gat-great.py',
             'scripts/gat-compare.py',
             'scripts/gat-plot.py'],
    zip_safe=False,
)
